#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy

import numpy as np
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.parameter import ChoiceParameter
from ax.modelbridge.transforms.metrics_as_task import MetricsAsTask
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_search_space_for_range_values


class MetricsAsTaskTransformTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.metric_task_map = {
            "metric1": ["metric2", "metric3"],
            "metric2": ["metric3"],
        }
        self.observations = [
            Observation(
                data=ObservationData(
                    metric_names=["metric1", "metric2", "metric3"],
                    means=np.array([1.0, 2.0, 3.0]),
                    covariance=np.diag([1.0, 2.0, 3.0]),
                ),
                features=ObservationFeatures(parameters={"x": 5.0, "y": 2.0}),
                arm_name="0_0",
            ),
            Observation(
                data=ObservationData(
                    metric_names=["metric3"],
                    means=np.array([30.0]),
                    covariance=np.array([[30.0]]),
                ),
                features=ObservationFeatures(parameters={"x": 10.0, "y": 4.0}),
            ),
        ]
        self.search_space = get_search_space_for_range_values(min=0.0, max=20.0)
        self.expected_new_observations = [
            Observation(
                data=ObservationData(
                    metric_names=["metric1", "metric2", "metric3"],
                    means=np.array([1.0, 2.0, 3.0]),
                    covariance=np.diag([1.0, 2.0, 3.0]),
                ),
                features=ObservationFeatures(
                    parameters={"x": 5.0, "y": 2.0, "METRIC_TASK": "TARGET"}
                ),
                arm_name="0_0",
            ),
            Observation(
                data=ObservationData(
                    metric_names=["metric2", "metric3"],
                    means=np.array([1.0, 1.0]),
                    covariance=np.diag([1.0, 1.0]),
                ),
                features=ObservationFeatures(
                    parameters={"x": 5.0, "y": 2.0, "METRIC_TASK": "metric1"}
                ),
                arm_name="0_0",
            ),
            Observation(
                data=ObservationData(
                    metric_names=["metric3"],
                    means=np.array([2.0]),
                    covariance=np.array([[2.0]]),
                ),
                features=ObservationFeatures(
                    parameters={"x": 5.0, "y": 2.0, "METRIC_TASK": "metric2"}
                ),
                arm_name="0_0",
            ),
            Observation(
                data=ObservationData(
                    metric_names=["metric3"],
                    means=np.array([30.0]),
                    covariance=np.array([[30.0]]),
                ),
                features=ObservationFeatures(
                    parameters={"x": 10.0, "y": 4.0, "METRIC_TASK": "TARGET"}
                ),
            ),
        ]
        self.t = MetricsAsTask(
            search_space=self.search_space,
            observations=self.observations,
            config={"metric_task_map": self.metric_task_map},
        )

    def test_Init(self) -> None:
        with self.assertRaises(ValueError):
            MetricsAsTask(
                search_space=self.search_space, observations=self.observations
            )

    def test_TransformObservations(self) -> None:
        new_obs = self.t.transform_observations(deepcopy(self.observations))
        self.assertEqual(new_obs, self.expected_new_observations)

        new_obs = self.t.untransform_observations(new_obs)
        self.assertEqual(new_obs, self.observations)

    def test_TransformObservationFeatures(self) -> None:
        obsfs_t = self.t.transform_observation_features(
            deepcopy([obs.features for obs in self.observations])
        )
        for obsf in obsfs_t:
            assert obsf.parameters["METRIC_TASK"] == "TARGET"

        obsfs_t = self.t.untransform_observation_features(obsfs_t)
        for obsf in obsfs_t:
            assert "METRIC_TASK" not in obsf.parameters

        with self.assertRaises(ValueError):
            self.t.untransform_observation_features(
                deepcopy([obs.features for obs in self.expected_new_observations])
            )

    def test_TransformSearchSpace(self) -> None:
        new_ss = self.t._transform_search_space(deepcopy(self.search_space))
        self.assertEqual(len(new_ss.parameters), 3)
        new_param = new_ss.parameters["METRIC_TASK"]
        self.assertIsInstance(new_param, ChoiceParameter)
        self.assertEqual(
            new_param.values, ["TARGET", "metric1", "metric2"]  # pyre-ignore
        )
        self.assertTrue(new_param.is_task)  # pyre-ignore
        self.assertEqual(new_param.target_value, "TARGET")



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import numpy as np
import torch
from ax.utils.common.testutils import TestCase
from ax.utils.testing.utils import generic_equals


class TestUtils(TestCase):
    def test_generic_equals(self) -> None:
        # Basics.
        self.assertTrue(generic_equals(5, 5))
        self.assertFalse(generic_equals(5, 1))
        self.assertTrue(generic_equals("abc", "abc"))
        self.assertFalse(generic_equals("abc", "abcd"))
        self.assertFalse(generic_equals("abc", 5))
        # With tensors.
        self.assertTrue(generic_equals(torch.ones(2), torch.ones(2)))
        self.assertFalse(generic_equals(torch.ones(2), torch.zeros(2)))
        self.assertFalse(generic_equals(torch.ones(2), [0, 0]))
        # Dictionaries.
        self.assertTrue(generic_equals({"a": torch.ones(2)}, {"a": torch.ones(2)}))
        self.assertFalse(generic_equals({"a": torch.ones(2)}, {"a": torch.zeros(2)}))
        self.assertFalse(
            generic_equals({"a": torch.ones(2)}, {"a": torch.ones(2), "b": 5})
        )
        self.assertFalse(generic_equals({"a": torch.ones(2)}, [torch.ones(2)]))
        self.assertTrue(
            generic_equals({"a": torch.ones(2), "b": 2}, {"b": 2, "a": torch.ones(2)})
        )
        # Tuple / list.
        self.assertTrue(generic_equals([3, 2], [3, 2]))
        self.assertTrue(generic_equals([3, (2, 3)], [3, (2, 3)]))
        self.assertFalse(generic_equals([3, (2, 3)], [3, (2, 4)]))
        self.assertFalse(generic_equals([0, 1], range(2)))
        # Other.
        self.assertTrue(generic_equals(range(2), range(2)))
        self.assertTrue(generic_equals(np.ones(2), np.ones(2)))
        self.assertFalse(generic_equals(np.ones(2), np.zeros(2)))
        self.assertTrue(generic_equals({1, 2}, {1, 2}))
        self.assertFalse(generic_equals({1, 2}, {1, 2, 3}))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from enum import Enum
from unittest import mock

import numpy as np
import pandas as pd
from ax.core.arm import Arm
from ax.core.generator_run import GeneratorRun
from ax.metrics.chemistry import ChemistryMetric, ChemistryProblemType
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_trial


class DummyEnum(Enum):
    DUMMY: str = "dummy"


class ChemistryMetricTest(TestCase):
    def test_ChemistryMetric(self) -> None:
        # basic test
        read_csv = pd.read_csv
        for problem_type in (
            ChemistryProblemType.DIRECT_ARYLATION,
            ChemistryProblemType.SUZUKI,
        ):
            with mock.patch(
                "ax.metrics.chemistry.pd.read_csv",
                wraps=lambda filename, index_col: read_csv(
                    filename, index_col=index_col, nrows=1
                ),
            ) as mock_read_csv:
                metric = ChemistryMetric(name="test_metric", problem_type=problem_type)
                self.assertFalse(metric.noiseless)
                self.assertIs(metric.problem_type, problem_type)
                self.assertFalse(metric.lower_is_better)
                if problem_type is ChemistryProblemType.DIRECT_ARYLATION:
                    param_names = [
                        "Base_SMILES",
                        "Concentration",
                        "Ligand_SMILES",
                        "Solvent_SMILES",
                        "Temp_C",
                    ]
                    param_values = (
                        "O=C([O-])C.[K+]",
                        0.1,
                        (
                            "CC(C)C1=CC(C(C)C)=C(C(C(C)C)=C1)C2=C(P(C3CCCCC3)"
                            "C4CCCCC4)C(OC)=CC=C2OC"
                        ),
                        "CC(N(C)C)=O",
                        105,
                    )
                    obj = 5.47
                else:
                    param_names = [
                        "Base_SMILES",
                        "Electrophile_SMILES",
                        "Ligand_SMILES",
                        "Nucleophile_SMILES",
                        "Solvent_SMILES",
                    ]
                    param_values = (
                        "[Na+].[OH-]",
                        "ClC1=CC=C(N=CC=C2)C2=C1",
                        "CC(P(C(C)(C)C)C(C)(C)C)(C)C",
                        "CC1=CC=C(N(C2CCCCO2)N=C3)C3=C1B(O)O",
                        "N#CC",
                    )
                    obj = 4.76

                params = dict(zip(param_names, param_values))
                trial = get_trial()
                trial._generator_run = GeneratorRun(
                    # pyre-fixme[6]: For 2nd argument expected `Dict[str,
                    #  Union[None, bool, float, int, str]]` but got `Dict[str,
                    #  Union[float, int, str]]`.
                    arms=[Arm(name="0_0", parameters=params)]
                )
                df = metric.fetch_trial_data(trial).unwrap().df
                self.assertEqual(mock_read_csv.call_count, 1)
                self.assertEqual(df["mean"].values[0], obj)
                self.assertTrue(np.isnan(df["sem"].values[0]))
                # test caching
                metric.fetch_trial_data(trial)
                self.assertEqual(mock_read_csv.call_count, 1)

                # test noiseless
                metric = ChemistryMetric(
                    name="test_metric", problem_type=problem_type, noiseless=True
                )
                df = metric.fetch_trial_data(trial).unwrap().df
                self.assertEqual(df["sem"].values[0], 0.0)



#!/usr/bin/env fbpython
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.core.utils import get_model_times
from ax.modelbridge.registry import Models
from ax.telemetry.experiment import ExperimentCompletedRecord, ExperimentCreatedRecord
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_experiment,
    get_experiment_with_custom_runner_and_metric,
)
from ax.utils.testing.mock import fast_botorch_optimize


class TestExperiment(TestCase):
    def test_experiment_created_record_from_experiment(self) -> None:
        experiment = get_experiment_with_custom_runner_and_metric(
            has_outcome_constraint=True
        )

        record = ExperimentCreatedRecord.from_experiment(experiment=experiment)
        expected = ExperimentCreatedRecord(
            experiment_name="test",
            experiment_type=None,
            num_continuous_range_parameters=1,
            num_int_range_parameters_small=0,
            num_int_range_parameters_medium=0,
            num_int_range_parameters_large=1,
            num_log_scale_range_parameters=0,
            num_unordered_choice_parameters_small=1,
            num_unordered_choice_parameters_medium=0,
            num_unordered_choice_parameters_large=0,
            num_fixed_parameters=1,
            dimensionality=3,
            hierarchical_tree_height=1,
            num_parameter_constraints=3,
            num_objectives=1,
            num_tracking_metrics=1,
            num_outcome_constraints=1,
            num_map_metrics=0,
            metric_cls_to_quantity={"Metric": 2, "CustomTestMetric": 1},
            runner_cls="CustomTestRunner",
        )
        self.assertEqual(record, expected)

    def test_experiment_completed_record_from_experiment(self) -> None:
        experiment = get_experiment_with_custom_runner_and_metric(
            has_outcome_constraint=True, num_trials=1
        )
        record = ExperimentCompletedRecord.from_experiment(experiment=experiment)

        # Calculate these here, may change from run to run
        fit_time, gen_time = get_model_times(experiment=experiment)
        expected = ExperimentCompletedRecord(
            num_initialization_trials=1,
            num_bayesopt_trials=0,
            num_other_trials=0,
            num_completed_trials=1,
            num_failed_trials=0,
            num_abandoned_trials=0,
            num_early_stopped_trials=0,
            total_fit_time=int(fit_time),
            total_gen_time=int(gen_time),
        )
        self.assertEqual(record, expected)

    @fast_botorch_optimize
    def test_bayesopt_trials_are_trials_containing_bayesopt(self) -> None:
        experiment = get_branin_experiment()
        sobol = Models.SOBOL(search_space=experiment.search_space)
        trial = experiment.new_batch_trial().add_generator_run(
            generator_run=sobol.gen(5)
        )
        trial.mark_completed(unsafe=True)

        # create a trial that among other things does bayesopt
        botorch = Models.BOTORCH_MODULAR(
            experiment=experiment, data=experiment.fetch_data()
        )
        trial = (
            experiment.new_batch_trial()
            .add_generator_run(generator_run=sobol.gen(2))
            .add_generator_run(generator_run=botorch.gen(5))
        )
        trial.add_arm(experiment.arms_by_name["0_0"])
        trial.mark_completed(unsafe=True)

        record = ExperimentCompletedRecord.from_experiment(experiment=experiment)
        self.assertEqual(record.num_initialization_trials, 1)
        self.assertEqual(record.num_bayesopt_trials, 1)
        self.assertEqual(record.num_other_trials, 0)

    def test_other_trials_are_trials_with_no_models(self) -> None:
        experiment = get_branin_experiment()
        sobol = Models.SOBOL(search_space=experiment.search_space)
        trial = experiment.new_batch_trial().add_generator_run(
            generator_run=sobol.gen(5)
        )
        trial.mark_completed(unsafe=True)

        # create a trial that has no GRs that used models
        trial = experiment.new_batch_trial()
        trial.add_arm(experiment.arms_by_name["0_0"])
        trial.mark_completed(unsafe=True)

        record = ExperimentCompletedRecord.from_experiment(experiment=experiment)
        self.assertEqual(record.num_initialization_trials, 1)
        self.assertEqual(record.num_bayesopt_trials, 0)
        self.assertEqual(record.num_other_trials, 1)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from contextlib import ExitStack
from typing import Any, cast, Dict, Tuple
from unittest import mock

import numpy as np
import torch
from ax.core.search_space import SearchSpaceDigest
from ax.models.torch.botorch_moo import MultiObjectiveBotorchModel
from ax.models.torch.botorch_moo_defaults import (
    get_outcome_constraint_transforms,
    get_qLogEHVI,
    get_qLogNEHVI,
    get_weighted_mc_objective_and_objective_thresholds,
    infer_objective_thresholds,
    pareto_frontier_evaluator,
)
from ax.models.torch.utils import _get_X_pending_and_observed
from ax.utils.common.random import with_rng_seed
from ax.utils.common.testutils import TestCase
from botorch.models.gp_regression import SingleTaskGP
from botorch.utils.datasets import SupervisedDataset
from botorch.utils.multi_objective.hypervolume import infer_reference_point
from botorch.utils.testing import MockModel, MockPosterior
from torch._tensor import Tensor


MOO_DEFAULTS_PATH: str = "ax.models.torch.botorch_moo_defaults"
GET_ACQF_PATH: str = MOO_DEFAULTS_PATH + ".get_acquisition_function"
GET_CONSTRAINT_PATH: str = MOO_DEFAULTS_PATH + ".get_outcome_constraint_transforms"
GET_OBJ_PATH: str = (
    MOO_DEFAULTS_PATH + ".get_weighted_mc_objective_and_objective_thresholds"
)
FIT_MODEL_MO_PATH = "ax.models.torch.botorch_defaults.fit_gpytorch_mll"


# pyre-fixme[2]: Parameter must be annotated.
def dummy_predict(model, X) -> Tuple[Tensor, Tensor]:
    # Add column to X that is a product of previous elements.
    mean = torch.cat([X, torch.prod(X, dim=1).reshape(-1, 1)], dim=1)
    cov = torch.zeros(mean.shape[0], mean.shape[1], mean.shape[1])
    return mean, cov


class FrontierEvaluatorTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.X = torch.tensor(
            [[1.0, 0.0], [1.0, 1.0], [1.0, 3.0], [2.0, 2.0], [3.0, 1.0]]
        )
        self.Y = torch.tensor(
            [
                [1.0, 0.0, 0.0],
                [1.0, 1.0, 1.0],
                [1.0, 3.0, 3.0],
                [2.0, 2.0, 4.0],
                [3.0, 1.0, 3.0],
            ]
        )
        self.Yvar = torch.zeros(5, 3)
        self.outcome_constraints = (
            torch.tensor([[0.0, 0.0, 1.0]]),
            torch.tensor([[3.5]]),
        )
        self.objective_thresholds = torch.tensor([0.5, 1.5])
        self.objective_weights = torch.tensor([1.0, 1.0])
        bounds = [(0.0, 4.0), (0.0, 4.0)]
        self.model = MultiObjectiveBotorchModel(model_predictor=dummy_predict)
        with mock.patch(FIT_MODEL_MO_PATH) as _mock_fit_model:
            self.model.fit(
                datasets=[
                    SupervisedDataset(
                        X=self.X,
                        Y=self.Y,
                        Yvar=self.Yvar,
                        feature_names=["x1", "x2"],
                        outcome_names=["a", "b", "c"],
                    )
                ],
                search_space_digest=SearchSpaceDigest(
                    feature_names=["x1", "x2"],
                    bounds=bounds,
                ),
            )
            _mock_fit_model.assert_called_once()

    def test_pareto_frontier_raise_error_when_missing_data(self) -> None:
        with self.assertRaises(ValueError):
            pareto_frontier_evaluator(
                model=self.model,
                objective_thresholds=self.objective_thresholds,
                objective_weights=self.objective_weights,
                Yvar=self.Yvar,
            )

    def test_pareto_frontier_evaluator_raw(self) -> None:
        Yvar = torch.diag_embed(self.Yvar)
        Y, cov, indx = pareto_frontier_evaluator(
            model=self.model,
            objective_weights=self.objective_weights,
            objective_thresholds=self.objective_thresholds,
            Y=self.Y,
            Yvar=Yvar,
        )
        pred = self.Y[2:4]
        self.assertTrue(torch.allclose(Y, pred), f"{Y} does not match {pred}")
        expected_cov = Yvar[2:4]
        self.assertTrue(torch.allclose(expected_cov, cov))
        self.assertTrue(torch.equal(torch.arange(2, 4), indx))

        # Omit objective_thresholds
        Y, cov, indx = pareto_frontier_evaluator(
            model=self.model,
            objective_weights=self.objective_weights,
            Y=self.Y,
            Yvar=Yvar,
        )
        pred = self.Y[2:]
        self.assertTrue(torch.allclose(Y, pred), f"{Y} does not match {pred}")
        expected_cov = Yvar[2:]
        self.assertTrue(torch.allclose(expected_cov, cov))
        self.assertTrue(torch.equal(torch.arange(2, 5), indx))

        # Change objective_weights so goal is to minimize b
        Y, cov, indx = pareto_frontier_evaluator(
            model=self.model,
            objective_weights=torch.tensor([1.0, -1.0]),
            objective_thresholds=self.objective_thresholds,
            Y=self.Y,
            Yvar=Yvar,
        )
        pred = self.Y[[0, 4]]
        self.assertTrue(
            torch.allclose(Y, pred), f"actual {Y} does not match pred {pred}"
        )
        expected_cov = Yvar[[0, 4]]
        self.assertTrue(torch.allclose(expected_cov, cov))

        # test no points better than reference point
        Y, cov, indx = pareto_frontier_evaluator(
            model=self.model,
            objective_weights=self.objective_weights,
            objective_thresholds=torch.full_like(self.objective_thresholds, 100.0),
            Y=self.Y,
            Yvar=Yvar,
        )
        self.assertTrue(torch.equal(Y, self.Y[:0]))
        self.assertTrue(torch.equal(cov, torch.zeros(0, 3, 3)))
        self.assertTrue(torch.equal(torch.tensor([], dtype=torch.long), indx))

    def test_pareto_frontier_evaluator_predict(self) -> None:
        Y, cov, indx = pareto_frontier_evaluator(
            model=self.model,
            objective_weights=self.objective_weights,
            objective_thresholds=self.objective_thresholds,
            X=self.X,
        )
        pred = self.Y[2:4]
        self.assertTrue(
            torch.allclose(Y, pred), f"actual {Y} does not match pred {pred}"
        )
        self.assertTrue(torch.equal(torch.arange(2, 4), indx))

    def test_pareto_frontier_evaluator_with_outcome_constraints(self) -> None:
        Y, cov, indx = pareto_frontier_evaluator(
            model=self.model,
            objective_weights=self.objective_weights,
            objective_thresholds=self.objective_thresholds,
            Y=self.Y,
            Yvar=self.Yvar,
            outcome_constraints=self.outcome_constraints,
        )
        pred = self.Y[2, :]
        self.assertTrue(
            torch.allclose(Y, pred), f"actual {Y} does not match pred {pred}"
        )
        self.assertTrue(torch.equal(torch.tensor([2], dtype=torch.long), indx))

    def test_pareto_frontier_evaluator_with_nan(self) -> None:
        Y = torch.cat([self.Y, torch.zeros(5, 1)], dim=-1)
        Yvar = torch.zeros(5, 4, 4)
        weights = torch.tensor([1.0, 1.0, 0.0, 0.0])
        outcome_constraints = (
            torch.tensor([[0.0, 0.0, 1.0, 0.0]]),
            torch.tensor([[3.5]]),
        )
        # Evaluate without NaNs as a baseline.
        _, _, idx = pareto_frontier_evaluator(
            model=None, objective_weights=weights, Y=Y, Yvar=Yvar
        )
        self.assertEqual(idx.tolist(), [2, 3, 4])
        # Set an element of idx 2 to NaN. Should be removed.
        Y[2, 1] = float("nan")
        _, _, idx = pareto_frontier_evaluator(
            model=None, objective_weights=weights, Y=Y, Yvar=Yvar
        )
        self.assertEqual(idx.tolist(), [3, 4])
        # Set the unused constraint element of idx 3 to NaN. No effect.
        Y[3, 2] = float("nan")
        _, _, idx = pareto_frontier_evaluator(
            model=None, objective_weights=weights, Y=Y, Yvar=Yvar
        )
        self.assertEqual(idx.tolist(), [3, 4])
        # Add constraint, 3 should be removed.
        _, _, idx = pareto_frontier_evaluator(
            model=None,
            objective_weights=weights,
            Y=Y,
            Yvar=Yvar,
            outcome_constraints=outcome_constraints,
        )
        self.assertEqual(idx.tolist(), [4])
        # Set unused index of 4 to NaN. No effect.
        Y[4, 3] = float("nan")
        _, _, idx = pareto_frontier_evaluator(
            model=None,
            objective_weights=weights,
            Y=Y,
            Yvar=Yvar,
            outcome_constraints=outcome_constraints,
        )
        self.assertEqual(idx.tolist(), [4])


class BotorchMOODefaultsTest(TestCase):
    def test_get_qLogEHVI_input_validation_errors(self) -> None:
        weights = torch.ones(2)
        objective_thresholds = torch.zeros(2)
        mm = MockModel(MockPosterior())
        with self.assertRaisesRegex(
            ValueError, "There are no feasible observed points."
        ):
            get_qLogEHVI(
                model=mm,
                objective_weights=weights,
                objective_thresholds=objective_thresholds,
            )

    def test_get_weighted_mc_objective_and_objective_thresholds(self) -> None:
        objective_weights = torch.tensor([0.0, 1.0, 0.0, 1.0])
        objective_thresholds = torch.arange(4, dtype=torch.float)
        (
            weighted_obj,
            new_obj_thresholds,
        ) = get_weighted_mc_objective_and_objective_thresholds(
            objective_weights=objective_weights,
            objective_thresholds=objective_thresholds,
        )
        self.assertTrue(torch.equal(weighted_obj.weights, objective_weights[[1, 3]]))
        self.assertEqual(weighted_obj.outcomes.tolist(), [1, 3])
        self.assertTrue(torch.equal(new_obj_thresholds, objective_thresholds[[1, 3]]))

    def test_get_qLogNEHVI_input_validation_errors(self) -> None:
        model = MultiObjectiveBotorchModel()
        weights = torch.ones(2)
        objective_thresholds = torch.zeros(2)
        with self.assertRaisesRegex(
            ValueError, "There are no feasible observed points."
        ):
            get_qLogNEHVI(
                # pyre-fixme[6]: For 1st param expected `Model` but got
                #  `Optional[Model]`.
                model=model._model,
                objective_weights=weights,
                objective_thresholds=objective_thresholds,
            )

    @mock.patch(  # pyre-ignore
        "ax.models.torch.botorch_moo_defaults._check_posterior_type",
        wraps=lambda y: y,
    )
    def test_get_qLogEHVI(self, _) -> None:
        weights = torch.tensor([0.0, 1.0, 1.0])
        X_observed = torch.rand(4, 3)
        X_pending = torch.rand(1, 3)
        constraints = (torch.tensor([1.0, 0.0, 0.0]), torch.tensor([[10.0]]))
        Y = torch.rand(4, 3)
        mm = MockModel(MockPosterior(mean=Y))
        objective_thresholds = torch.arange(3, dtype=torch.float)
        obj_and_obj_t = get_weighted_mc_objective_and_objective_thresholds(
            objective_weights=weights,
            objective_thresholds=objective_thresholds,
        )
        (weighted_obj, new_obj_thresholds) = obj_and_obj_t
        cons_tfs = get_outcome_constraint_transforms(constraints)
        with with_rng_seed(0):
            seed = torch.randint(1, 10000, (1,)).item()
        with ExitStack() as es:
            mock_get_acqf = es.enter_context(mock.patch(GET_ACQF_PATH))
            es.enter_context(
                mock.patch(MOO_DEFAULTS_PATH + ".checked_cast", wraps=cast)
            )
            es.enter_context(mock.patch(GET_CONSTRAINT_PATH, return_value=cons_tfs))
            es.enter_context(mock.patch(GET_OBJ_PATH, return_value=obj_and_obj_t))
            es.enter_context(with_rng_seed(0))
            get_qLogEHVI(
                model=mm,
                objective_weights=weights,
                outcome_constraints=constraints,
                objective_thresholds=objective_thresholds,
                X_observed=X_observed,
                X_pending=X_pending,
            )
            mock_get_acqf.assert_called_once_with(
                acquisition_function_name="qLogEHVI",
                model=mm,
                objective=weighted_obj,
                X_observed=X_observed,
                X_pending=X_pending,
                constraints=cons_tfs,
                mc_samples=128,
                alpha=0.0,
                seed=seed,
                ref_point=new_obj_thresholds.tolist(),
                Y=Y,
            )

    # test infer objective thresholds alone
    @mock.patch(  # pyre-ignore
        "ax.models.torch.botorch_moo_defaults._check_posterior_type",
        wraps=lambda y: y,
    )
    def test_infer_objective_thresholds(self, _, cuda: bool = False) -> None:
        # TODO: refactor this test into smaller test cases
        for dtype in (torch.float, torch.double):
            tkwargs: Dict[str, Any] = {
                "device": torch.device("cuda") if cuda else torch.device("cpu"),
                "dtype": dtype,
            }
            Xs = [torch.tensor([[1.0, 2.0, 3.0], [2.0, 3.0, 4.0]], **tkwargs)]
            bounds = [(0.0, 1.0), (1.0, 4.0), (2.0, 5.0)]
            outcome_constraints = (
                torch.tensor([[1.0, 0.0, 0.0]], **tkwargs),
                torch.tensor([[10.0]], **tkwargs),
            )
            linear_constraints = (
                torch.tensor([1.0, 0.0, 0.0], **tkwargs),
                torch.tensor([2.0], **tkwargs),
            )
            objective_weights = torch.tensor([-1.0, -1.0, 0.0], **tkwargs)
            with ExitStack() as es:
                _mock_get_X_pending_and_observed = es.enter_context(
                    mock.patch(
                        "ax.models.torch.botorch_moo_defaults."
                        "_get_X_pending_and_observed",
                        wraps=_get_X_pending_and_observed,
                    )
                )
                _mock_infer_reference_point = es.enter_context(
                    mock.patch(
                        "ax.models.torch.botorch_moo_defaults.infer_reference_point",
                        wraps=infer_reference_point,
                    )
                )
                model = SingleTaskGP(train_X=Xs[0], train_Y=torch.rand(2, 3, **tkwargs))
                es.enter_context(
                    mock.patch.object(
                        model,
                        "posterior",
                        return_value=MockPosterior(
                            mean=torch.tensor(
                                [
                                    [11.0, 2.0],
                                    [9.0, 3.0],
                                ],
                                **tkwargs,
                            )
                        ),
                    )
                )

                # test passing Xs
                obj_thresholds = infer_objective_thresholds(
                    model,
                    bounds=bounds,
                    objective_weights=objective_weights,
                    outcome_constraints=outcome_constraints,
                    fixed_features={},
                    linear_constraints=linear_constraints,
                    Xs=Xs + Xs,
                )
                _mock_get_X_pending_and_observed.assert_called_once()
                ckwargs = _mock_get_X_pending_and_observed.call_args[1]
                actual_Xs = ckwargs["Xs"]
                for X in actual_Xs:
                    self.assertTrue(torch.equal(X, Xs[0]))
                self.assertEqual(ckwargs["bounds"], bounds)
                self.assertTrue(
                    torch.equal(ckwargs["objective_weights"], objective_weights)
                )
                oc = ckwargs["outcome_constraints"]
                self.assertTrue(torch.equal(oc[0], outcome_constraints[0]))
                self.assertTrue(torch.equal(oc[1], outcome_constraints[1]))
                self.assertEqual(ckwargs["fixed_features"], {})
                lc = ckwargs["linear_constraints"]
                self.assertTrue(torch.equal(lc[0], linear_constraints[0]))
                self.assertTrue(torch.equal(lc[1], linear_constraints[1]))
                _mock_infer_reference_point.assert_called_once()
                ckwargs = _mock_infer_reference_point.call_args[1]
                self.assertEqual(ckwargs["scale"], 0.1)
                self.assertTrue(
                    torch.equal(
                        ckwargs["pareto_Y"],
                        torch.tensor([[-9.0, -3.0]], **tkwargs),
                    )
                )
                self.assertTrue(
                    torch.equal(obj_thresholds[:2], torch.tensor([9.9, 3.3], **tkwargs))
                )
                self.assertTrue(np.isnan(obj_thresholds[2].item()))
            with ExitStack() as es:
                _mock_get_X_pending_and_observed = es.enter_context(
                    mock.patch(
                        "ax.models.torch.botorch_moo_defaults."
                        "_get_X_pending_and_observed",
                        wraps=_get_X_pending_and_observed,
                    )
                )
                _mock_infer_reference_point = es.enter_context(
                    mock.patch(
                        "ax.models.torch.botorch_moo_defaults.infer_reference_point",
                        wraps=infer_reference_point,
                    )
                )
                es.enter_context(
                    mock.patch.object(
                        model,
                        "posterior",
                        return_value=MockPosterior(
                            mean=torch.tensor(
                                [
                                    [11.0, 2.0],
                                    [9.0, 3.0],
                                ],
                                **tkwargs,
                            )
                        ),
                    )
                )

                # test passing X_observed
                obj_thresholds = infer_objective_thresholds(
                    model,
                    objective_weights=objective_weights,
                    outcome_constraints=outcome_constraints,
                    X_observed=Xs[0],
                )
                _mock_get_X_pending_and_observed.assert_not_called()
                self.assertTrue(
                    torch.equal(obj_thresholds[:2], torch.tensor([9.9, 3.3], **tkwargs))
                )
                self.assertTrue(np.isnan(obj_thresholds[2].item()))
            # test that value error is raised if bounds are not supplied
            with self.assertRaises(ValueError):
                infer_objective_thresholds(
                    model,
                    objective_weights=objective_weights,
                    Xs=Xs + Xs,
                )
            # test that value error is raised if Xs are not supplied
            with self.assertRaises(ValueError):
                infer_objective_thresholds(
                    model,
                    bounds=bounds,
                    objective_weights=objective_weights,
                )
            # test subset_model without subset_idcs
            es.enter_context(
                mock.patch.object(
                    model,
                    "posterior",
                    return_value=MockPosterior(
                        mean=torch.tensor(
                            [
                                [11.0, 2.0],
                                [9.0, 3.0],
                            ],
                            **tkwargs,
                        )
                    ),
                )
            )

            obj_thresholds = infer_objective_thresholds(
                model,
                objective_weights=objective_weights,
                outcome_constraints=outcome_constraints,
                X_observed=Xs[0],
            )
            self.assertTrue(
                torch.equal(obj_thresholds[:2], torch.tensor([9.9, 3.3], **tkwargs))
            )
            self.assertTrue(np.isnan(obj_thresholds[2].item()))
            # test passing subset_idcs
            subset_idcs = torch.tensor(
                [0, 1],
                dtype=torch.long,
                device=tkwargs["device"],
            )
            obj_thresholds = infer_objective_thresholds(
                model,
                objective_weights=objective_weights,
                outcome_constraints=outcome_constraints,
                X_observed=Xs[0],
                subset_idcs=subset_idcs,
            )
            self.assertTrue(
                torch.equal(obj_thresholds[:2], torch.tensor([9.9, 3.3], **tkwargs))
            )
            self.assertTrue(np.isnan(obj_thresholds[2].item()))
            # test without subsetting (e.g. if there are
            # 3 metrics for 2 objectives + 1 outcome constraint)
            outcome_constraints = (
                torch.tensor([[0.0, 0.0, 1.0]], **tkwargs),
                torch.tensor([[5.0]], **tkwargs),
            )
            with ExitStack() as es:
                _mock_get_X_pending_and_observed = es.enter_context(
                    mock.patch(
                        "ax.models.torch.botorch_moo_defaults."
                        "_get_X_pending_and_observed",
                        wraps=_get_X_pending_and_observed,
                    )
                )
                _mock_infer_reference_point = es.enter_context(
                    mock.patch(
                        "ax.models.torch.botorch_moo_defaults.infer_reference_point",
                        wraps=infer_reference_point,
                    )
                )
                es.enter_context(
                    mock.patch.object(
                        model,
                        "posterior",
                        return_value=MockPosterior(
                            mean=torch.tensor(
                                [
                                    [11.0, 2.0, 6.0],
                                    [9.0, 3.0, 4.0],
                                ],
                                **tkwargs,
                            )
                        ),
                    )
                )
                # test passing Xs
                obj_thresholds = infer_objective_thresholds(
                    model,
                    bounds=bounds,
                    objective_weights=objective_weights,
                    outcome_constraints=outcome_constraints,
                    fixed_features={},
                    linear_constraints=linear_constraints,
                    Xs=Xs + Xs + Xs,
                )
                self.assertTrue(
                    torch.equal(obj_thresholds[:2], torch.tensor([9.9, 3.3], **tkwargs))
                )
                self.assertTrue(np.isnan(obj_thresholds[2].item()))

    def test_infer_objective_thresholds_cuda(self) -> None:
        if torch.cuda.is_available():
            self.test_infer_objective_thresholds(cuda=True)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import dataclasses
from collections import OrderedDict
from copy import deepcopy
from itertools import product
from typing import Dict, Type
from unittest import mock
from unittest.mock import Mock

import numpy as np
import torch
from ax.core.search_space import SearchSpaceDigest
from ax.exceptions.core import AxWarning, UnsupportedError
from ax.models.torch.botorch_modular.acquisition import Acquisition
from ax.models.torch.botorch_modular.model import (
    BoTorchModel,
    choose_botorch_acqf_class,
    SurrogateSpec,
)
from ax.models.torch.botorch_modular.surrogate import Surrogate
from ax.models.torch.botorch_modular.utils import (
    choose_model_class,
    construct_acquisition_and_optimizer_options,
)
from ax.models.torch.utils import _filter_X_observed
from ax.models.torch_base import TorchOptConfig
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import checked_cast
from ax.utils.testing.mock import fast_botorch_optimize
from ax.utils.testing.torch_stubs import get_torch_test_data
from botorch.acquisition import qLogNoisyExpectedImprovement
from botorch.acquisition.input_constructors import (
    _register_acqf_input_constructor,
    get_acqf_input_constructor,
)
from botorch.acquisition.monte_carlo import qExpectedImprovement
from botorch.acquisition.multi_objective.logei import (
    qLogNoisyExpectedHypervolumeImprovement,
)
from botorch.acquisition.multi_objective.objective import WeightedMCMultiOutputObjective
from botorch.acquisition.objective import GenericMCObjective
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP
from botorch.models.gp_regression import SingleTaskGP
from botorch.models.gp_regression_fidelity import SingleTaskMultiFidelityGP
from botorch.models.model import Model, ModelList
from botorch.sampling.normal import SobolQMCNormalSampler
from botorch.utils.constraints import get_outcome_constraint_transforms
from botorch.utils.datasets import SupervisedDataset
from gpytorch.likelihoods.gaussian_likelihood import FixedNoiseGaussianLikelihood
from gpytorch.mlls.exact_marginal_log_likelihood import ExactMarginalLogLikelihood
from pyre_extensions import none_throws


CURRENT_PATH: str = __name__
MODEL_PATH: str = BoTorchModel.__module__
SURROGATE_PATH: str = Surrogate.__module__
ACQUISITION_PATH: str = Acquisition.__module__

ACQ_OPTIONS: Dict[str, SobolQMCNormalSampler] = {
    Keys.SAMPLER: SobolQMCNormalSampler(sample_shape=torch.Size([1024]))
}


class BoTorchModelTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.botorch_model_class = SingleTaskGP
        self.surrogate = Surrogate(botorch_model_class=self.botorch_model_class)
        self.acquisition_class = Acquisition
        self.botorch_acqf_class = qExpectedImprovement
        self.acquisition_options = ACQ_OPTIONS
        self.model = BoTorchModel(
            surrogate=self.surrogate,
            acquisition_class=self.acquisition_class,
            botorch_acqf_class=self.botorch_acqf_class,
            acquisition_options=self.acquisition_options,
        )

        self.dtype = torch.float
        self.device = torch.device("cpu")
        tkwargs = {"dtype": self.dtype, "device": self.device}
        self.tkwargs = tkwargs
        (
            self.Xs,
            self.Ys,
            self.Yvars,
            self.bounds,
            _,
            self.feature_names,  # This is ["x1", "x2", "x3"].
            self.metric_names,  # This is just ["y"].
        ) = get_torch_test_data(dtype=self.dtype)
        Xs2, Ys2, Yvars2, _, _, _, _ = get_torch_test_data(dtype=self.dtype, offset=1.0)
        self.X_test = Xs2[0]
        self.block_design_training_data = [
            SupervisedDataset(
                X=self.Xs[0],
                Y=self.Ys[0],
                Yvar=self.Yvars[0],
                feature_names=self.feature_names,
                outcome_names=self.metric_names,
            )
        ]
        self.non_block_design_training_data = self.block_design_training_data + [
            SupervisedDataset(
                X=Xs2[0],
                Y=Ys2[0],
                Yvar=Yvars2[0],
                feature_names=self.feature_names,
                outcome_names=["y2"],
            )
        ]
        self.metric_names_for_list_surrogate = ["y", "y2"]
        self.search_space_digest = SearchSpaceDigest(
            feature_names=self.feature_names,
            bounds=[(0.0, 10.0), (0.0, 10.0), (0.0, 10.0)],
        )
        self.mf_search_space_digest = SearchSpaceDigest(
            feature_names=self.feature_names,
            bounds=[(0.0, 10.0), (0.0, 10.0), (0.0, 10.0)],
            fidelity_features=[2],
            target_values={1: 1.0},
        )
        self.candidate_metadata = []
        self.optimizer_options = {Keys.NUM_RESTARTS: 40, Keys.RAW_SAMPLES: 1024}
        self.model_gen_options = {Keys.OPTIMIZER_KWARGS: self.optimizer_options}
        self.objective_weights = torch.tensor([1.0], **tkwargs)
        self.outcome_constraints = (
            torch.tensor([[1.0]], **tkwargs),
            torch.tensor([[-5.0]], **tkwargs),
        )
        self.moo_objective_weights = torch.tensor([1.0, 1.5, 0.0], **tkwargs)
        self.moo_objective_thresholds = torch.tensor(
            [0.5, 1.5, float("nan")], **tkwargs
        )
        self.moo_outcome_constraints = (
            torch.tensor([[1.0, 0.0, 0.0]], **tkwargs),
            torch.tensor([[-5.0]], **tkwargs),
        )
        self.linear_constraints = None
        self.fixed_features = None
        self.pending_observations = None
        self.moo_metric_names = ["y1", "y2", "y3"]
        self.moo_training_data = [  # block design
            SupervisedDataset(
                X=X,
                Y=Y,
                Yvar=Yvar,
                feature_names=self.feature_names,
                outcome_names=[mn],
            )
            for X, Y, Yvar, mn in zip(
                checked_cast(list, self.Xs) * 3,
                self.Ys + Ys2 + self.Ys,
                checked_cast(list, self.Yvars) * 3,
                self.moo_metric_names,
            )
        ]

        self.torch_opt_config = TorchOptConfig(
            objective_weights=self.objective_weights,
            outcome_constraints=self.outcome_constraints,
            linear_constraints=self.linear_constraints,
            fixed_features=self.fixed_features,
            pending_observations=self.pending_observations,
            model_gen_options=self.model_gen_options,
        )
        self.moo_torch_opt_config = dataclasses.replace(
            self.torch_opt_config,
            objective_weights=self.moo_objective_weights,
            objective_thresholds=self.moo_objective_thresholds,
            outcome_constraints=self.moo_outcome_constraints,
        )

    def test_init(self) -> None:
        # Default model with no specifications.
        model = BoTorchModel()
        self.assertEqual(model.acquisition_class, Acquisition)
        # Model that specifies `botorch_acqf_class`.
        model = BoTorchModel(botorch_acqf_class=qExpectedImprovement)
        self.assertEqual(model.acquisition_class, Acquisition)
        self.assertEqual(model.botorch_acqf_class, qExpectedImprovement)

        # Check defaults for refitting settings.
        self.assertFalse(model.refit_on_cv)
        self.assertTrue(model.warm_start_refit)

        # Check setting non-default refitting settings
        mdl2 = BoTorchModel(
            surrogate=self.surrogate,
            acquisition_class=self.acquisition_class,
            acquisition_options=self.acquisition_options,
            refit_on_cv=True,
            warm_start_refit=False,
        )
        self.assertTrue(mdl2.refit_on_cv)
        self.assertFalse(mdl2.warm_start_refit)

    def test_surrogates_property(self) -> None:
        self.assertEqual(self.surrogate, list(self.model.surrogates.values())[0])

    def test_Xs_property(self) -> None:
        self.model.fit(
            datasets=self.block_design_training_data,
            search_space_digest=self.search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )

        self.assertEqual(len(self.model.Xs), 1)
        self.assertTrue(
            self.model.Xs[0].equal(torch.tensor([[1.0, 2.0, 3.0], [2.0, 3.0, 4.0]]))
        )

        with self.assertRaisesRegex(NotImplementedError, "Xs not implemented"):
            self.model._surrogates = {"foo": Surrogate(), "bar": Surrogate()}
            self.model.Xs

    def test_dtype(self) -> None:
        self.model.fit(
            datasets=self.block_design_training_data,
            search_space_digest=self.search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )
        self.assertEqual(self.model.dtype, torch.float32)

    def test_device(self) -> None:
        self.model.fit(
            datasets=self.block_design_training_data,
            search_space_digest=self.search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )
        self.assertEqual(self.model.device, torch.device("cpu"))

    def test_botorch_acqf_class_property(self) -> None:
        self.assertEqual(self.botorch_acqf_class, self.model.botorch_acqf_class)
        self.model._botorch_acqf_class = None
        with self.assertRaisesRegex(
            ValueError, "BoTorch `AcquisitionFunction` has not yet been set."
        ):
            self.model.botorch_acqf_class

    @mock.patch(f"{SURROGATE_PATH}.use_model_list", return_value=False)
    def test__construct__raises_on_mixed_data(self, _: Mock) -> None:
        """Ensure proper error is raised when mixing data w/ and w/o variance."""
        ds1, ds2 = self.non_block_design_training_data
        datasets = [
            ds1,
            SupervisedDataset(
                X=ds2.X,
                Y=ds2.Y,
                feature_names=ds2.feature_names,
                outcome_names=ds2.outcome_names,
            ),
        ]
        with self.assertRaisesRegex(
            UnsupportedError, "Cannot convert mixed data with and without variance"
        ):
            self.model.fit(
                datasets=datasets,
                search_space_digest=self.search_space_digest,
            )

    @mock.patch(f"{SURROGATE_PATH}.use_model_list", return_value=False)
    def test__construct__converts_non_block(self, _: Mock) -> None:
        """Ensure non-block design data is converted with warnings."""
        ds = self.block_design_training_data[0]
        X1 = ds.X
        X2 = torch.cat((X1[:1], torch.rand_like(X1[1:])))
        datasets = [
            ds,
            SupervisedDataset(
                X=X2,
                Y=ds.Y,
                Yvar=ds.Yvar,
                feature_names=ds.feature_names,
                outcome_names=self.metric_names_for_list_surrogate[1:],
            ),
        ]
        with self.assertWarnsRegex(
            AxWarning, "Forcing converion of data not complying to a block design"
        ):
            self.model.fit(
                datasets=datasets,
                search_space_digest=self.search_space_digest,
            )

    def test__construct(self) -> None:
        """Test autoset."""
        self.model._surrogates = {}
        with mock.patch(
            f"{SURROGATE_PATH}.choose_model_class", wraps=choose_model_class
        ) as mock_choose_model_class:
            self.model.fit(
                datasets=self.block_design_training_data,
                search_space_digest=self.mf_search_space_digest,
                candidate_metadata=self.candidate_metadata,
            )
        # `choose_model_class` is called.
        mock_choose_model_class.assert_called_with(
            datasets=self.block_design_training_data,
            search_space_digest=self.mf_search_space_digest,
        )

    @mock.patch(f"{SURROGATE_PATH}.Surrogate._construct_model")
    def test_fit(self, mock_fit: Mock) -> None:
        # If surrogate is not yet set, initialize it with dispatcher functions.
        self.model._surrogates = {}
        with self.assertRaisesRegex(RuntimeError, "is not initialized. Must `fit`"):
            self.model.search_space_digest  # can't access before fit

        with self.assertRaisesRegex(RuntimeError, "manually is disallowed"):
            self.model.search_space_digest = self.mf_search_space_digest

        self.model.fit(
            datasets=self.block_design_training_data,
            search_space_digest=self.mf_search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )

        self.assertIsInstance(self.model.search_space_digest, SearchSpaceDigest)
        self.assertEqual(self.model.search_space_digest, self.mf_search_space_digest)
        self.assertEqual(self.model.output_order, [0])

        # Since we want to refit on updates but not warm start refit, we clear the
        # state dict.
        mock_fit.assert_called_with(
            dataset=self.block_design_training_data[0],
            search_space_digest=self.mf_search_space_digest,
            botorch_model_class=SingleTaskMultiFidelityGP,
            state_dict=None,
            refit=True,
        )

        labels_to_outcomes = self.model.outcomes_by_surrogate_label
        self.assertIsInstance(labels_to_outcomes, dict)
        self.assertEqual(len(labels_to_outcomes), 1)
        self.assertEqual(next(iter(labels_to_outcomes.values())), self.metric_names)

    @mock.patch(f"{SURROGATE_PATH}.Surrogate.predict")
    def test_predict(self, mock_predict: Mock) -> None:
        self.model.predict(X=self.X_test)
        mock_predict.assert_called_with(X=self.X_test)

    @fast_botorch_optimize
    def test_multiple_surrogates(self) -> None:
        surrogate_specs = {
            "Vanilla": SurrogateSpec(
                botorch_model_class=SingleTaskGP,
                outcomes=["y1", "y3"],
            ),
            "Bayesian": SurrogateSpec(
                botorch_model_class=SaasFullyBayesianSingleTaskGP,
                outcomes=["y2"],
            ),
        }
        model = BoTorchModel(surrogate_specs=surrogate_specs)
        model.fit(
            datasets=self.moo_training_data,
            search_space_digest=self.search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )
        self.assertEqual(list(model.surrogates.keys()), list(surrogate_specs.keys()))
        self.assertEqual(model.output_order, [0, 2, 1])
        f1, cov1 = model.surrogates["Vanilla"].predict(self.X_test)
        f2, cov2 = model.surrogates["Bayesian"].predict(self.X_test)
        n = self.X_test.shape[0]
        true_f = torch.zeros(n, 3, dtype=self.X_test.dtype, device=self.X_test.device)
        true_f[:, 0] = f1[:, 0]
        true_f[:, 2] = f1[:, 1]
        true_f[:, 1] = f2[:, 0]
        true_cov = torch.zeros(
            n, 3, 3, dtype=self.X_test.dtype, device=self.X_test.device
        )
        for i in range(2):
            true_cov[i, 0, 0] = cov1[i, 0, 0]
            true_cov[i, 2, 2] = cov1[i, 1, 1]
            true_cov[i, 1, 1] = cov2[i, 0, 0]
        f_joint, cov_joint = model.predict(self.X_test)
        self.assertTrue(torch.allclose(true_f, f_joint))
        self.assertTrue(torch.allclose(true_cov, cov_joint))

    @mock.patch(f"{MODEL_PATH}.BoTorchModel.fit")
    def test_cross_validate(self, mock_fit: Mock) -> None:
        self.model.fit(
            datasets=self.block_design_training_data,
            search_space_digest=self.mf_search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )

        old_surrogate = self.model.surrogates[Keys.ONLY_SURROGATE]
        old_surrogate._model = mock.MagicMock()
        old_surrogate._model.state_dict.return_value = OrderedDict({"key": "val"})

        for refit_on_cv, warm_start_refit, use_posterior_predictive in product(
            (True, False), (True, False), (True, False)
        ):
            self.model.refit_on_cv = refit_on_cv
            self.model.warm_start_refit = warm_start_refit
            with mock.patch(
                f"{SURROGATE_PATH}.Surrogate.clone_reset",
                return_value=mock.MagicMock(spec=Surrogate),
            ) as mock_clone_reset:
                self.model.cross_validate(
                    datasets=self.block_design_training_data,
                    X_test=self.X_test,
                    search_space_digest=self.mf_search_space_digest,
                    use_posterior_predictive=use_posterior_predictive,
                )
                # Check that `predict` is called on the cloned surrogate, not
                # on the original one.
                mock_predict = mock_clone_reset.return_value.predict
                mock_predict.assert_called_once()

                # Check correct X_test.
                self.assertTrue(
                    torch.equal(
                        mock_predict.call_args_list[-1][1].get("X"),
                        self.X_test,
                    ),
                )
                self.assertIs(
                    mock_predict.call_args_list[-1][1]["use_posterior_predictive"],
                    use_posterior_predictive,
                )

            # Check that surrogate is reset back to `old_surrogate` at the
            # end of cross-validation.
            self.assertTrue(self.model.surrogates[Keys.ONLY_SURROGATE] is old_surrogate)

            expected_state_dict = (
                None
                if refit_on_cv and not warm_start_refit
                else self.model.surrogates[Keys.ONLY_SURROGATE].model.state_dict()
            )

            # Check correct `refit` and `state_dict` values.
            self.assertEqual(mock_fit.call_args_list[-1][1].get("refit"), refit_on_cv)
            if expected_state_dict is None:
                self.assertIsNone(
                    mock_fit.call_args_list[-1][1].get("state_dict"),
                    expected_state_dict,
                )
            else:
                self.assertEqual(
                    mock_fit.call_args_list[-1][1]
                    .get("state_dicts")
                    .get(Keys.ONLY_SURROGATE)
                    .keys(),
                    expected_state_dict.keys(),
                )

    @fast_botorch_optimize
    @mock.patch(
        f"{MODEL_PATH}.construct_acquisition_and_optimizer_options",
        wraps=construct_acquisition_and_optimizer_options,
    )
    @mock.patch(f"{CURRENT_PATH}.Acquisition.optimize")
    @mock.patch(
        f"{MODEL_PATH}.choose_botorch_acqf_class", wraps=choose_botorch_acqf_class
    )
    def _test_gen(
        self,
        mock_choose_botorch_acqf_class: Mock,
        mock_optimize: Mock,
        mock_construct_options: Mock,
        botorch_model_class: Type[Model],
        search_space_digest: SearchSpaceDigest,
    ) -> None:
        qLogNEI_input_constructor = get_acqf_input_constructor(
            qLogNoisyExpectedImprovement
        )
        mock_input_constructor = mock.MagicMock(
            qLogNEI_input_constructor, side_effect=qLogNEI_input_constructor
        )
        _register_acqf_input_constructor(
            acqf_cls=qLogNoisyExpectedImprovement,
            input_constructor=mock_input_constructor,
        )
        mock_optimize.return_value = (
            torch.tensor([[1.0]]),
            torch.tensor([2.0]),
            torch.tensor([1.0]),
        )
        surrogate = Surrogate(botorch_model_class=botorch_model_class)
        model = BoTorchModel(
            surrogate=surrogate,
            acquisition_class=Acquisition,
            acquisition_options=self.acquisition_options,
        )
        model.surrogates[Keys.ONLY_SURROGATE].fit(
            datasets=self.block_design_training_data,
            search_space_digest=search_space_digest,
        )
        model._botorch_acqf_class = None
        # Assert that error is raised if we haven't fit the model
        with self.assertRaises(RuntimeError):
            model.gen(
                n=1,
                search_space_digest=search_space_digest,
                torch_opt_config=self.torch_opt_config,
            )
        # Add search space digest reference to make the model think it's been fit
        model._search_space_digest = search_space_digest
        with mock.patch.object(
            BoTorchModel,
            "_instantiate_acquisition",
            wraps=model._instantiate_acquisition,
        ) as mock_init_acqf:
            model.gen(
                n=1,
                search_space_digest=search_space_digest,
                torch_opt_config=self.torch_opt_config,
            )
        # Assert acquisition initialized with expected arguments
        mock_init_acqf.assert_called_once_with(
            search_space_digest=search_space_digest,
            torch_opt_config=self.torch_opt_config,
            acq_options=self.acquisition_options,
        )

        mock_input_constructor.assert_called_once()
        ckwargs = mock_input_constructor.call_args[1]

        # We particularly want to make sure that args that will not be used are
        # not passed
        expected_kwargs = {
            "bounds",
            "constraints",
            "X_baseline",
            "sampler",
            "objective",
            "training_data",
            "model",
        }
        self.assertSetEqual(set(ckwargs.keys()), expected_kwargs)
        for k in expected_kwargs:
            self.assertIsNotNone(ckwargs[k], f"{k} is None")

        m = ckwargs["model"]
        self.assertIsInstance(m, botorch_model_class)
        self.assertEqual(m.num_outputs, 1)
        training_data = ckwargs["training_data"]
        self.assertIsInstance(training_data, SupervisedDataset)
        self.assertTrue(torch.equal(training_data.X, self.Xs[0]))
        self.assertTrue(
            torch.equal(
                training_data.Y,
                torch.cat([ds.Y for ds in self.block_design_training_data], dim=-1),
            )
        )

        self.assertIsInstance(ckwargs["objective"], GenericMCObjective)
        expected_X_baseline = _filter_X_observed(
            Xs=[dataset.X for dataset in self.block_design_training_data],
            objective_weights=self.objective_weights,
            outcome_constraints=self.outcome_constraints,
            bounds=search_space_digest.bounds,
            linear_constraints=self.linear_constraints,
            fixed_features=self.fixed_features,
        )
        self.assertTrue(
            torch.equal(ckwargs["X_baseline"], none_throws(expected_X_baseline))
        )

        # Assert `construct_acquisition_and_optimizer_options` called with kwargs
        mock_construct_options.assert_called_with(
            acqf_options=self.acquisition_options,
            model_gen_options=self.model_gen_options,
        )
        # Assert `choose_botorch_acqf_class` is called
        mock_choose_botorch_acqf_class.assert_called_once()
        self.assertEqual(model._botorch_acqf_class, qLogNoisyExpectedImprovement)

        # Assert `optimize` called with kwargs
        mock_optimize.assert_called_with(
            n=1,
            search_space_digest=search_space_digest,
            inequality_constraints=None,
            fixed_features=self.fixed_features,
            rounding_func=None,
            optimizer_options=self.optimizer_options,
        )

        _register_acqf_input_constructor(
            acqf_cls=qLogNoisyExpectedImprovement,
            input_constructor=qLogNEI_input_constructor,
        )

    def test_gen_SingleTaskGP(self) -> None:
        self._test_gen(
            botorch_model_class=SingleTaskGP,
            search_space_digest=self.search_space_digest,
        )

    def test_gen_SingleTaskMultiFidelityGP(self) -> None:
        self._test_gen(
            botorch_model_class=SingleTaskMultiFidelityGP,
            search_space_digest=self.mf_search_space_digest,
        )

    @fast_botorch_optimize
    def test_feature_importances(self) -> None:
        for botorch_model_class in [SingleTaskGP, SaasFullyBayesianSingleTaskGP]:
            surrogate = Surrogate(botorch_model_class=botorch_model_class)
            model = BoTorchModel(
                surrogate=surrogate,
                acquisition_class=Acquisition,
                acquisition_options=self.acquisition_options,
            )
            model.surrogates[Keys.ONLY_SURROGATE].fit(
                datasets=self.block_design_training_data,
                search_space_digest=SearchSpaceDigest(feature_names=[], bounds=[]),
            )
            if botorch_model_class == SaasFullyBayesianSingleTaskGP:
                mcmc_samples = {
                    "lengthscale": torch.tensor(
                        [[1, 2, 3], [2, 3, 4], [3, 4, 5]], **self.tkwargs
                    ),
                    "outputscale": torch.rand(3, **self.tkwargs),
                    "mean": torch.randn(3, **self.tkwargs),
                    "noise": torch.rand(3, **self.tkwargs),
                }
                model.surrogate.model.load_mcmc_samples(mcmc_samples)
                importances = model.feature_importances()
                self.assertTrue(
                    np.allclose(importances, np.array([6 / 13, 4 / 13, 3 / 13]))
                )
                self.assertEqual(importances.shape, (1, 1, 3))
                saas_model = deepcopy(model.surrogate.model)
            else:
                model.surrogate.model.covar_module.base_kernel.lengthscale = (
                    torch.tensor([1, 2, 3], **self.tkwargs)
                )
                importances = model.feature_importances()
                self.assertTrue(
                    np.allclose(importances, np.array([6 / 11, 3 / 11, 2 / 11]))
                )
                self.assertEqual(importances.shape, (1, 1, 3))
                vanilla_model = deepcopy(model.surrogate.model)

        # Mixed model
        model.surrogate._model = ModelList(saas_model, vanilla_model)  # pyre-ignore
        importances = model.feature_importances()
        self.assertTrue(
            np.allclose(
                importances,
                np.expand_dims(
                    np.array([[6 / 13, 4 / 13, 3 / 13], [6 / 11, 3 / 11, 2 / 11]]),
                    axis=1,
                ),
            )
        )
        self.assertEqual(importances.shape, (2, 1, 3))
        # Add model we don't support
        vanilla_model.covar_module.base_kernel = None
        model.surrogate._model = vanilla_model  # pyre-ignore
        with self.assertRaisesRegex(
            NotImplementedError,
            "Failed to extract lengthscales from `m.covar_module.base_kernel`",
        ):
            model.feature_importances()
        # Test model is None
        model.surrogate._model = None
        with self.assertRaisesRegex(
            ValueError, "BoTorch `Model` has not yet been constructed"
        ):
            model.feature_importances()
        # Test unsupported surrogate
        model._surrogates = {"vanilla": None, "saas": None}
        with self.assertRaisesRegex(
            NotImplementedError,
            "feature_importances not implemented for multi-surrogate case",
        ):
            model.feature_importances()

    @fast_botorch_optimize
    def test_best_point(self) -> None:
        self.model._surrogates = {}
        self.model.fit(
            datasets=self.block_design_training_data,
            search_space_digest=self.mf_search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )
        self.assertIsNotNone(
            self.model.best_point(
                search_space_digest=self.mf_search_space_digest,
                torch_opt_config=self.torch_opt_config,
            )
        )
        with mock.patch(f"{SURROGATE_PATH}.best_in_sample_point", return_value=None):
            self.assertIsNone(
                self.model.best_point(
                    search_space_digest=self.mf_search_space_digest,
                    torch_opt_config=self.torch_opt_config,
                )
            )
        with self.assertRaisesRegex(NotImplementedError, "Best observed"):
            self.model.best_point(
                search_space_digest=self.mf_search_space_digest,
                torch_opt_config=dataclasses.replace(
                    self.torch_opt_config, is_moo=True
                ),
            )

    @fast_botorch_optimize
    @mock.patch(
        f"{MODEL_PATH}.construct_acquisition_and_optimizer_options",
        return_value=({"num_fantasies": 64}, {"num_restarts": 40, "raw_samples": 1024}),
    )
    @mock.patch(f"{CURRENT_PATH}.Acquisition", autospec=True)
    def test_evaluate_acquisition_function(
        self, mock_acquisition: Mock, _mock_construct_options: Mock
    ) -> None:
        model = BoTorchModel(
            surrogate=self.surrogate,
            acquisition_class=Acquisition,
            acquisition_options=self.acquisition_options,
        )
        model.surrogates[Keys.ONLY_SURROGATE].fit(
            datasets=self.block_design_training_data,
            search_space_digest=SearchSpaceDigest(
                feature_names=[],
                bounds=[],
            ),
        )
        model.evaluate_acquisition_function(
            X=self.X_test,
            search_space_digest=self.search_space_digest,
            torch_opt_config=self.torch_opt_config,
            acq_options=self.acquisition_options,
        )
        # testing that the new setup chooses qLogNEI by default
        self.assertEqual(model._botorch_acqf_class, qLogNoisyExpectedImprovement)
        # `mock_acquisition` is a mock of the Acquisition class, so to check the mock's
        # `evaluate` on an instance of that class, we use
        # `mock_acquisition.return_value.evaluate`.
        mock_acquisition.return_value.evaluate.assert_called()

    @mock.patch(f"{MODEL_PATH}.Surrogate", wraps=Surrogate)
    @mock.patch(f"{SURROGATE_PATH}.Surrogate._construct_model", return_value=None)
    @mock.patch(f"{SURROGATE_PATH}.ModelListGP")
    def test_surrogate_model_options_propagation(
        self, _m1: Mock, _m2: Mock, mock_init: Mock
    ) -> None:
        model = BoTorchModel(
            surrogate_specs={
                "name": SurrogateSpec(
                    botorch_model_kwargs={"some_option": "some_value"}
                )
            }
        )
        model.fit(
            datasets=self.non_block_design_training_data,
            search_space_digest=self.mf_search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )
        mock_init.assert_called_with(
            botorch_model_class=None,
            model_options={"some_option": "some_value"},
            mll_class=ExactMarginalLogLikelihood,
            mll_options={},
            covar_module_class=None,
            covar_module_options=None,
            likelihood_class=None,
            likelihood_options=None,
            input_transform_classes=None,
            input_transform_options=None,
            outcome_transform_classes=None,
            outcome_transform_options=None,
            allow_batched_models=True,
        )

    @mock.patch(f"{MODEL_PATH}.Surrogate", wraps=Surrogate)
    @mock.patch(f"{SURROGATE_PATH}.Surrogate._construct_model", return_value=None)
    @mock.patch(f"{SURROGATE_PATH}.ModelListGP")
    def test_surrogate_options_propagation(
        self, _m1: Mock, _m2: Mock, mock_init: Mock
    ) -> None:
        model = BoTorchModel(
            surrogate_specs={"name": SurrogateSpec(allow_batched_models=False)}
        )
        model.fit(
            datasets=self.non_block_design_training_data,
            search_space_digest=self.mf_search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )
        mock_init.assert_called_with(
            botorch_model_class=None,
            model_options={},
            mll_class=ExactMarginalLogLikelihood,
            mll_options={},
            covar_module_class=None,
            covar_module_options=None,
            likelihood_class=None,
            likelihood_options=None,
            input_transform_classes=None,
            input_transform_options=None,
            outcome_transform_classes=None,
            outcome_transform_options=None,
            allow_batched_models=False,
        )

    @mock.patch(
        f"{ACQUISITION_PATH}.Acquisition.optimize",
        # Dummy candidates and acquisition function value.
        return_value=(torch.tensor([[2.0]]), torch.tensor([1.0])),
    )
    def test_model_list_choice(self, _) -> None:  # , mock_extract_training_data):
        model = BoTorchModel()
        model.fit(
            datasets=self.non_block_design_training_data,
            search_space_digest=self.mf_search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )
        self.assertEqual(model.output_order, [0, 1])
        # A model list should be chosen, since Xs are not all the same.
        model_list = checked_cast(
            ModelList, model.surrogates[Keys.AUTOSET_SURROGATE].model
        )
        for submodel in model_list.models:
            # There are fidelity features and nonempty Yvars, so
            # MFGP should be chosen.
            self.assertIsInstance(submodel, SingleTaskMultiFidelityGP)

    @mock.patch(
        f"{ACQUISITION_PATH}.Acquisition.optimize",
        # Dummy candidates, acquisition value, and weights
        return_value=(torch.tensor([[2.0]]), torch.tensor([1.0]), torch.tensor([1.0])),
    )
    def test_MOO(self, _) -> None:
        # Add mock for qLogNEHVI input constructor to catch arguments passed to it.
        qLogNEHVI_input_constructor = get_acqf_input_constructor(
            qLogNoisyExpectedHypervolumeImprovement
        )
        mock_input_constructor = mock.MagicMock(
            qLogNEHVI_input_constructor, side_effect=qLogNEHVI_input_constructor
        )
        _register_acqf_input_constructor(
            acqf_cls=qLogNoisyExpectedHypervolumeImprovement,
            input_constructor=mock_input_constructor,
        )

        model = BoTorchModel()
        model.fit(
            datasets=self.moo_training_data,
            search_space_digest=self.search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )
        self.assertIsInstance(
            model.surrogates[Keys.AUTOSET_SURROGATE].model, SingleTaskGP
        )
        subset_outcome_constraints = (
            # model is subset since last output is not used
            self.moo_outcome_constraints[0][:, :2],
            self.moo_outcome_constraints[1],
        )
        constraints = get_outcome_constraint_transforms(
            outcome_constraints=subset_outcome_constraints,
        )
        with mock.patch(
            f"{ACQUISITION_PATH}.get_outcome_constraint_transforms",
            # Dummy candidates and acquisition function value.
            return_value=constraints,
        ) as mock_get_outcome_constraint_transforms:
            gen_results = model.gen(
                n=1,
                search_space_digest=self.mf_search_space_digest,
                torch_opt_config=self.moo_torch_opt_config,
            )
        mock_get_outcome_constraint_transforms.assert_called_once()
        ckwargs = mock_get_outcome_constraint_transforms.call_args[1]
        oc = ckwargs["outcome_constraints"]
        self.assertTrue(torch.equal(oc[0], subset_outcome_constraints[0]))
        self.assertTrue(torch.equal(oc[1], subset_outcome_constraints[1]))

        # Check input constructor args
        ckwargs = mock_input_constructor.call_args[1]
        expected_kwargs = {
            "constraints",
            "bounds",
            "objective",
            "training_data",
            "X_baseline",
            "model",
            "objective_thresholds",
        }
        self.assertSetEqual(set(ckwargs.keys()), expected_kwargs)
        for k in expected_kwargs:
            self.assertIsNotNone(ckwargs[k], f"{k} is None")

        self.assertIs(model.botorch_acqf_class, qLogNoisyExpectedHypervolumeImprovement)
        mock_input_constructor.assert_called_once()
        m = ckwargs["model"]
        self.assertIsInstance(m, SingleTaskGP)
        self.assertIsInstance(m.likelihood, FixedNoiseGaussianLikelihood)
        self.assertEqual(m.num_outputs, 2)
        training_data = ckwargs["training_data"]
        self.assertIsNotNone(training_data.Yvar)
        self.assertTrue(torch.equal(training_data.X, self.Xs[0]))
        self.assertTrue(
            torch.equal(
                training_data.Y,
                torch.cat([ds.Y for ds in self.moo_training_data], dim=-1),
            )
        )
        self.assertTrue(
            torch.equal(
                training_data.Yvar,
                torch.cat([ds.Yvar for ds in self.moo_training_data], dim=-1),
            )
        )
        self.assertTrue(
            torch.equal(
                ckwargs["objective_thresholds"], self.moo_objective_thresholds[:2]
            )
        )
        self.assertIs(ckwargs["constraints"], constraints)

        obj_t = gen_results.gen_metadata["objective_thresholds"]
        self.assertTrue(torch.equal(obj_t[:2], self.moo_objective_thresholds[:2]))
        self.assertTrue(np.isnan(obj_t[2].item()))

        self.assertIsInstance(ckwargs["objective"], WeightedMCMultiOutputObjective)
        self.assertTrue(
            torch.equal(
                ckwargs["objective"].weights,
                self.moo_objective_weights[:2],
            )
        )
        expected_X_baseline = _filter_X_observed(
            Xs=[dataset.X for dataset in self.moo_training_data],
            objective_weights=self.moo_objective_weights,
            outcome_constraints=self.moo_outcome_constraints,
            bounds=self.search_space_digest.bounds,
            linear_constraints=self.linear_constraints,
            fixed_features=self.fixed_features,
        )
        self.assertTrue(
            torch.equal(ckwargs["X_baseline"], none_throws(expected_X_baseline))
        )
        # test inferred objective_thresholds
        objective_weights = torch.tensor([-1.0, -1.0, 0.0])
        outcome_constraints = (
            torch.tensor([[1.0, 0.0, 0.0]]),
            torch.tensor([[10.0]]),
        )
        linear_constraints = (
            torch.tensor([[1.0, 0.0, 0.0]]),
            torch.tensor([[2.0]]),
        )

        torch_opt_config = dataclasses.replace(
            self.moo_torch_opt_config,
            objective_weights=objective_weights,
            objective_thresholds=None,
            outcome_constraints=outcome_constraints,
            linear_constraints=linear_constraints,
        )

        with mock.patch(
            "ax.models.torch.botorch_modular.acquisition." "infer_objective_thresholds",
            return_value=torch.tensor([9.9, 3.3, float("nan")]),
        ) as _mock_model_infer_objective_thresholds:
            gen_results = model.gen(
                n=1,
                search_space_digest=self.mf_search_space_digest,
                torch_opt_config=torch_opt_config,
            )
        expected_X_baseline = _filter_X_observed(
            Xs=[dataset.X for dataset in self.moo_training_data],
            objective_weights=objective_weights,
            outcome_constraints=outcome_constraints,
            bounds=self.search_space_digest.bounds,
            linear_constraints=linear_constraints,
            fixed_features=self.fixed_features,
        )
        ckwargs = _mock_model_infer_objective_thresholds.call_args[1]
        self.assertTrue(
            torch.equal(
                ckwargs["objective_weights"],
                objective_weights,
            )
        )
        oc = ckwargs["outcome_constraints"]
        self.assertTrue(torch.equal(oc[0], outcome_constraints[0]))
        self.assertTrue(torch.equal(oc[1], outcome_constraints[1]))
        m = ckwargs["model"]
        self.assertIsInstance(m, SingleTaskGP)
        self.assertIsInstance(m.likelihood, FixedNoiseGaussianLikelihood)
        self.assertEqual(m.num_outputs, 2)
        self.assertIn("objective_thresholds", gen_results.gen_metadata)
        obj_t = gen_results.gen_metadata["objective_thresholds"]
        self.assertTrue(torch.equal(obj_t[:2], torch.tensor([9.9, 3.3])))
        self.assertTrue(np.isnan(obj_t[2].item()))

        # Avoid polluting the registry for other tests; re-register correct input
        # contructor for qLogNEHVI.
        _register_acqf_input_constructor(
            acqf_cls=qLogNoisyExpectedHypervolumeImprovement,
            input_constructor=qLogNEHVI_input_constructor,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import pandas as pd
from ax.core.data import Data
from ax.core.map_data import MapData, MapKeyInfo
from ax.utils.common.testutils import TestCase


class MapDataTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.df = pd.DataFrame(
            [
                {
                    "arm_name": "0_0",
                    "epoch": 0,
                    "mean": 2.0,
                    "sem": 0.2,
                    "trial_index": 1,
                    "metric_name": "a",
                },
                {
                    "arm_name": "0_0",
                    "epoch": 0,
                    "mean": 1.8,
                    "sem": 0.3,
                    "trial_index": 1,
                    "metric_name": "b",
                },
                {
                    "arm_name": "0_1",
                    "epoch": 0,
                    "mean": 4.0,
                    "sem": 0.6,
                    "trial_index": 1,
                    "metric_name": "a",
                },
                {
                    "arm_name": "0_1",
                    "epoch": 0,
                    "mean": 3.7,
                    "sem": 0.5,
                    "trial_index": 1,
                    "metric_name": "b",
                },
                {
                    "arm_name": "0_1",
                    "epoch": 1,
                    "mean": 0.5,
                    "sem": None,
                    "trial_index": 1,
                    "metric_name": "a",
                },
                {
                    "arm_name": "0_1",
                    "epoch": 1,
                    "mean": 3.0,
                    "sem": None,
                    "trial_index": 1,
                    "metric_name": "b",
                },
            ]
        )

        self.map_key_infos = [
            MapKeyInfo(
                key="epoch",
                default_value=0,
            )
        ]

        self.mmd = MapData(df=self.df, map_key_infos=self.map_key_infos)

    def test_map_key_info(self) -> None:
        self.assertEqual(self.map_key_infos, self.mmd.map_key_infos)

        self.assertEqual(self.mmd.map_key_infos[0].key, "epoch")
        self.assertEqual(self.mmd.map_key_infos[0].default_value, 0)
        self.assertEqual(self.mmd.map_key_infos[0].value_type, int)

    def test_init(self) -> None:
        empty = MapData()
        self.assertTrue(empty.map_df.empty)

        with self.assertRaisesRegex(ValueError, "map_key_infos may be `None` iff"):
            MapData(df=self.df, map_key_infos=None)

    def test_properties(self) -> None:
        self.assertEqual(self.mmd.map_key_infos, self.map_key_infos)
        self.assertEqual(self.mmd.map_keys, ["epoch"])
        self.assertEqual(self.mmd.map_key_to_type, {"epoch": int})

    def test_combine(self) -> None:
        data = MapData.from_multiple_map_data([])
        self.assertEqual(data.map_df.size, 0)

        mmd_double = MapData.from_multiple_map_data([self.mmd, self.mmd])
        self.assertEqual(mmd_double.map_df.size, 2 * self.mmd.map_df.size)
        self.assertEqual(mmd_double.map_key_infos, self.mmd.map_key_infos)

        different_map_df = pd.DataFrame(
            [
                {
                    "arm_name": "0_3",
                    "timestamp": 11,
                    "mean": 2.0,
                    "sem": 0.2,
                    "trial_index": 1,
                    "metric_name": "a",
                },
                {
                    "arm_name": "0_3",
                    "timestamp": 18,
                    "mean": 1.8,
                    "sem": 0.3,
                    "trial_index": 1,
                    "metric_name": "b",
                },
            ]
        )
        different_map_key_infos = [MapKeyInfo(key="timestamp", default_value=0.0)]
        different_mmd = MapData(
            df=different_map_df, map_key_infos=different_map_key_infos
        )

        combined = MapData.from_multiple_map_data([self.mmd, different_mmd])
        self.assertEqual(
            len(combined.map_df), len(self.mmd.map_df) + len(different_mmd.map_df)
        )
        self.assertEqual(combined.map_df.columns.size, self.mmd.map_df.columns.size + 1)
        self.assertEqual(
            combined.map_key_infos, self.map_key_infos + different_map_key_infos
        )

        combined_subset = MapData.from_multiple_map_data(
            [self.mmd, different_mmd], ["a"]
        )
        self.assertTrue((combined_subset.map_df["metric_name"] == "a").all())

        data_df = pd.DataFrame(
            [
                {
                    "arm_name": "0_4",
                    "mean": 2.0,
                    "sem": 0.2,
                    "trial_index": 1,
                    "metric_name": "a",
                },
                {
                    "arm_name": "0_4",
                    "mean": 1.8,
                    "sem": 0.3,
                    "trial_index": 1,
                    "metric_name": "b",
                },
            ]
        )
        data = Data(df=data_df)

        downcast_combined = MapData.from_multiple_data([self.mmd, data])
        self.assertEqual(
            len(downcast_combined.map_df), len(self.mmd.map_df) + len(data.df)
        )
        self.assertEqual(
            downcast_combined.map_df.columns.size, self.mmd.map_df.columns.size
        )
        self.assertEqual(downcast_combined.map_key_infos, self.map_key_infos)

        # Check that the Data's rows' epoch cell has the correct default value
        self.assertTrue(
            (
                downcast_combined.map_df[downcast_combined.map_df["arm_name"] == "0_4"][
                    "epoch"
                ]
                == self.mmd.map_key_infos[0].default_value
            ).all()
        )

    def test_from_map_evaluations(self) -> None:
        for sem in (0.5, None):
            eval1 = (3.7, sem) if sem is not None else 3.7
            eval2 = (3.8, sem) if sem is not None else 3.8
            map_data = MapData.from_map_evaluations(
                evaluations={
                    "0_1": [
                        ({"f1": 1.0, "f2": 0.5}, {"b": eval1}),
                        ({"f1": 1.0, "f2": 0.75}, {"b": eval2}),
                    ]
                },
                trial_index=0,
            )
            self.assertEqual(map_data.map_df["sem"].isnull().all(), sem is None)
            self.assertEqual(len(map_data.map_df), 2)
            self.assertEqual(set(map_data.map_keys), {"f1", "f2"})

        with self.assertRaisesRegex(
            ValueError, "Inconsistent map_key sets in evaluations"
        ):
            MapData.from_map_evaluations(
                evaluations={
                    "0_1": [
                        ({"f1": 1.0, "f2": 0.5}, {"b": (3.7, 0.5)}),
                    ]
                },
                map_key_infos=[MapKeyInfo(key="f1", default_value=0.0)],
                trial_index=0,
            )

    def test_upcast(self) -> None:
        fresh = MapData(df=self.df, map_key_infos=self.map_key_infos)
        self.assertIsNone(fresh._memo_df)  # Assert df is not cached before first call

        self.assertEqual(
            fresh.df.columns.size,
            fresh.map_df.columns.size - len(self.mmd.map_key_infos),
        )

        self.assertIsNotNone(fresh._memo_df)  # Assert df is cached after first call

    def test_subsample(self) -> None:
        arm_names = ["0_0", "1_0", "2_0", "3_0"]
        max_epochs = [25, 50, 75, 100]
        metric_names = ["a", "b"]
        large_map_df = pd.DataFrame(
            [
                {
                    "arm_name": arm_name,
                    "epoch": epoch + 1,
                    "mean": epoch * 0.1,
                    "sem": 0.1,
                    "trial_index": trial_index,
                    "metric_name": metric_name,
                }
                for metric_name in metric_names
                for trial_index, (arm_name, max_epoch) in enumerate(
                    zip(arm_names, max_epochs)
                )
                for epoch in range(max_epoch)
            ]
        )
        large_map_data = MapData(df=large_map_df, map_key_infos=self.map_key_infos)

        large_map_df_sparse_metric = pd.DataFrame(
            [
                {
                    "arm_name": arm_name,
                    "epoch": epoch + 1,
                    "mean": epoch * 0.1,
                    "sem": 0.1,
                    "trial_index": trial_index,
                    "metric_name": metric_name,
                }
                for metric_name in metric_names
                for trial_index, (arm_name, max_epoch) in enumerate(
                    zip(arm_names, max_epochs)
                )
                for epoch in range(max_epoch if metric_name == "a" else max_epoch // 5)
            ]
        )
        large_map_data_sparse_metric = MapData(
            df=large_map_df_sparse_metric, map_key_infos=self.map_key_infos
        )

        # test keep_every
        subsample = large_map_data.subsample(keep_every=10)
        self.assertEqual(len(subsample.map_df), 52)
        subsample = large_map_data.subsample(keep_every=25)
        self.assertEqual(len(subsample.map_df), 20)
        subsample = large_map_data.subsample(limit_rows_per_group=7)
        self.assertEqual(len(subsample.map_df), 36)
        # test passing specific map_key
        subsample = large_map_data.subsample(map_key="epoch", keep_every=10)
        self.assertEqual(len(subsample.map_df), 52)

        # test limit_rows_per_group
        subsample = large_map_data.subsample(limit_rows_per_group=1)
        self.assertEqual(len(subsample.map_df), 8)
        subsample = large_map_data.subsample(limit_rows_per_group=7)
        self.assertEqual(len(subsample.map_df), 36)
        subsample = large_map_data.subsample(limit_rows_per_group=10)
        self.assertEqual(len(subsample.map_df), 52)
        subsample = large_map_data.subsample(limit_rows_per_group=1000)
        self.assertEqual(len(subsample.map_df), 500)

        # test limit_rows_per_metric
        subsample = large_map_data.subsample(limit_rows_per_metric=50)
        self.assertEqual(len(subsample.map_df), 100)
        subsample = large_map_data.subsample(limit_rows_per_metric=65)
        self.assertEqual(len(subsample.map_df), 128)
        subsample = large_map_data.subsample(limit_rows_per_metric=1000)
        self.assertEqual(len(subsample.map_df), 500)

        # test include_first_last
        subsample = large_map_data.subsample(
            limit_rows_per_metric=20, include_first_last=True
        )
        self.assertEqual(len(subsample.map_df), 40)
        # check that we 1 and 100 are included
        self.assertEqual(subsample.map_df["epoch"].min(), 1)
        self.assertEqual(subsample.map_df["epoch"].max(), 100)
        subsample = large_map_data.subsample(
            limit_rows_per_metric=20, include_first_last=False
        )
        self.assertEqual(len(subsample.map_df), 40)
        self.assertEqual(subsample.map_df["epoch"].min(), 1)
        self.assertEqual(subsample.map_df["epoch"].max(), 92)

        # test limit_rows_per_metric when some metrics are sparsely
        # reported (we shouldn't subsample those)
        subsample = large_map_data_sparse_metric.subsample(
            limit_rows_per_metric=100, include_first_last=False
        )
        map_df = large_map_data_sparse_metric.map_df
        subsample_map_df = subsample.map_df
        self.assertEqual(
            len(subsample_map_df[subsample_map_df["metric_name"] == "a"]), 85
        )
        self.assertEqual(
            len(subsample_map_df[subsample_map_df["metric_name"] == "b"]),
            len(map_df[map_df["metric_name"] == "b"]),
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from math import ceil
from typing import Any, cast, Dict, List, Optional, Type

from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun
from ax.core.metric import Metric
from ax.core.trial import Trial
from ax.exceptions.core import ExperimentNotFoundError, ObjectNotFoundError
from ax.modelbridge.generation_strategy import GenerationStrategy
from ax.storage.sqa_store.db import session_scope
from ax.storage.sqa_store.decoder import Decoder
from ax.storage.sqa_store.reduced_state import (
    get_query_options_to_defer_immutable_duplicates,
    get_query_options_to_defer_large_model_cols,
)
from ax.storage.sqa_store.sqa_classes import (
    SQAExperiment,
    SQAGenerationStrategy,
    SQAGeneratorRun,
    SQAMetric,
    SQATrial,
)
from ax.storage.sqa_store.sqa_config import SQAConfig

from ax.storage.utils import MetricIntent
from ax.utils.common.constants import Keys
from ax.utils.common.typeutils import checked_cast, not_none
from sqlalchemy.orm import defaultload, lazyload, noload
from sqlalchemy.orm.exc import DetachedInstanceError


# ---------------------------- Loading `Experiment`. ---------------------------


def load_experiment(
    experiment_name: str,
    config: Optional[SQAConfig] = None,
    reduced_state: bool = False,
    load_trials_in_batches_of_size: Optional[int] = None,
    skip_runners_and_metrics: bool = False,
) -> Experiment:
    """Load experiment by name.

    Args:
        experiment_name: Name of the expeirment to load.
        config: `SQAConfig`, from which to retrieve the decoder. Optional,
            defaults to base `SQAConfig`.
        reduced_state: Whether to load experiment with a slightly reduced state
            (without abandoned arms on experiment and withoug model state,
            search space, and optimization config on generator runs).
        skip_runners_and_metrics: If True skip loading runners, and do only a
            minimal load of metrics. This option is intended to enable loading of
            experiments that require custom runners or metrics, without depending
            on a registry. Note that even though the intention is to skip loading
            of metrics, this option converts the loaded metrics into a base
            metric avoiding conversion related to custom properties of the metric.
    """
    config = config or SQAConfig()
    decoder = Decoder(config=config)
    return _load_experiment(
        experiment_name=experiment_name,
        decoder=decoder,
        reduced_state=reduced_state,
        load_trials_in_batches_of_size=load_trials_in_batches_of_size,
        skip_runners_and_metrics=skip_runners_and_metrics,
    )


def _load_experiment(
    experiment_name: str,
    decoder: Decoder,
    reduced_state: bool = False,
    load_trials_in_batches_of_size: Optional[int] = None,
    ax_object_field_overrides: Optional[Dict[str, Any]] = None,
    skip_runners_and_metrics: bool = False,
) -> Experiment:
    """Load experiment by name, using given Decoder instance.

    1) Get SQLAlchemy object from DB.
    2) Convert to corresponding Ax object.

    Args:
        experiment_name: Name of the experiment
        decoder: Decoder used to convert SQAlchemy objects into Ax objects
        reduced_state: Whether to load experiment and generation strategy
        load_trials_in_batches_of_size: Number of trials to be fetched from database
            per batch
        ax_object_field_overrides: Mapping of object types to mapping of fields
            to override values loaded objects will all be instantiated with fields
            set to override value
            current valid object types are: "runner"

    """

    # pyre-ignore Incompatible variable type [9]: exp_sqa_class is declared to have type
    # `Type[SQAExperiment]` but is used as type `Type[ax.storage.sqa_store.db.SQABase]`
    exp_sqa_class: Type[SQAExperiment] = decoder.config.class_to_sqa_class[Experiment]
    # pyre-ignore Incompatible variable type [9]: trial_sqa_class is decl. to have type
    # `Type[SQATrial]` but is used as type `Type[ax.storage.sqa_store.db.SQABase]`
    trial_sqa_class: Type[SQATrial] = decoder.config.class_to_sqa_class[Trial]
    imm_OC_and_SS = _get_experiment_immutable_opt_config_and_search_space(
        experiment_name=experiment_name, exp_sqa_class=exp_sqa_class
    )

    if reduced_state:
        _get_experiment_sqa_func = _get_experiment_sqa_reduced_state

    else:
        _get_experiment_sqa_func = (
            _get_experiment_sqa_immutable_opt_config_and_search_space
            if imm_OC_and_SS
            else _get_experiment_sqa
        )

    experiment_sqa = _get_experiment_sqa_func(
        experiment_name=experiment_name,
        exp_sqa_class=exp_sqa_class,
        trial_sqa_class=trial_sqa_class,
        load_trials_in_batches_of_size=load_trials_in_batches_of_size,
        skip_runners_and_metrics=skip_runners_and_metrics,
    )

    # If skip_runners_and_metrics is True, override metric types to
    # base metric type of 0 (to load them with minimum base properties).
    # This includes metrics attached to the experiment, and ones
    # attached to the trials.
    #
    # NOTE: Currently, we load metrics and then convert mostly go get
    # "lower_is_better". Alternatively, we can "nolod" all attributes except
    # "lower_is_better" or any other attribute we want to include. This can be
    # implemented in the future if we need to.
    if skip_runners_and_metrics:
        base_metric_type_int = decoder.config.metric_registry[Metric]
        for sqa_metric in experiment_sqa.metrics:
            _set_sqa_metric_to_base_type(
                sqa_metric, base_metric_type_int=base_metric_type_int
            )

        assign_metric_on_gr = not reduced_state and not imm_OC_and_SS
        if assign_metric_on_gr:
            try:
                for sqa_trial in experiment_sqa.trials:
                    for sqa_generator_run in sqa_trial.generator_runs:
                        for sqa_metric in sqa_generator_run.metrics:
                            _set_sqa_metric_to_base_type(
                                sqa_metric, base_metric_type_int=base_metric_type_int
                            )
            except DetachedInstanceError as e:
                raise DetachedInstanceError(
                    "Unable to retrieve metric from SQA generator run, possibly due "
                    "to parts of the experiment being lazy-loaded. This is not "
                    f"expected state, please contact Ax support. Original error: {e}"
                )

    return decoder.experiment_from_sqa(
        experiment_sqa=experiment_sqa,
        reduced_state=reduced_state,
        ax_object_field_overrides=ax_object_field_overrides,
    )


def _get_experiment_sqa(
    experiment_name: str,
    exp_sqa_class: Type[SQAExperiment],
    trial_sqa_class: Type[SQATrial],
    # pyre-fixme[2]: Parameter annotation cannot contain `Any`.
    trials_query_options: Optional[List[Any]] = None,
    load_trials_in_batches_of_size: Optional[int] = None,
    skip_runners_and_metrics: bool = False,
) -> SQAExperiment:
    """Obtains SQLAlchemy experiment object from DB."""
    with session_scope() as session:
        query = (
            session.query(exp_sqa_class).filter_by(name=experiment_name)
            # Delay loading trials to a separate call to `_get_trials_sqa` below
            .options(noload("trials"))
        )

        if skip_runners_and_metrics:
            query = query.options(noload("runners")).options(noload("trials.runner"))

        sqa_experiment = query.one_or_none()

    if sqa_experiment is None:
        raise ExperimentNotFoundError(f"Experiment '{experiment_name}' not found.")

    sqa_trials = _get_trials_sqa(
        experiment_id=sqa_experiment.id,
        trial_sqa_class=trial_sqa_class,
        trials_query_options=trials_query_options,
        load_trials_in_batches_of_size=load_trials_in_batches_of_size,
        skip_runners_and_metrics=skip_runners_and_metrics,
    )

    sqa_experiment.trials = sqa_trials

    return sqa_experiment


def _get_trials_sqa(
    experiment_id: int,
    trial_sqa_class: Type[SQATrial],
    load_trials_in_batches_of_size: Optional[int] = None,
    # pyre-fixme[2]: Parameter annotation cannot contain `Any`.
    trials_query_options: Optional[List[Any]] = None,
    skip_runners_and_metrics: bool = False,
) -> List[SQATrial]:
    """Obtains SQLAlchemy trial objects for given experiment ID from DB,
    optionally in mini-batches and with specified query options.
    """
    with session_scope() as session:
        query = session.query(trial_sqa_class.id).filter_by(experiment_id=experiment_id)
        trial_db_ids = query.all()
        trial_db_ids = [db_id_tuple[0] for db_id_tuple in trial_db_ids]

    if len(trial_db_ids) == 0:
        return []

    batch_size = (
        len(trial_db_ids)
        if load_trials_in_batches_of_size is None
        else load_trials_in_batches_of_size
    )

    sqa_trials = []
    for i in range(ceil(len(trial_db_ids) / batch_size)):
        mini_batch_db_ids = trial_db_ids[batch_size * i : batch_size * (i + 1)]
        with session_scope() as session:
            query = session.query(trial_sqa_class).filter(
                trial_sqa_class.id.in_(mini_batch_db_ids)  # pyre-ignore[16]
            )

            if trials_query_options is not None:
                query = query.options(*trials_query_options)

            if skip_runners_and_metrics:
                query = query.options(noload("runner"))

            sqa_trials.extend(query.all())

    return sqa_trials


def _set_sqa_metric_to_base_type(
    sqa_metric: SQAMetric, base_metric_type_int: int
) -> None:
    """Sets metric type to base type, since we don't want to load
    the metric class from the DB.
    """
    sqa_metric.metric_type = base_metric_type_int
    # Handle multi-objective metrics that are not directly attached to
    # the experiment
    if sqa_metric.intent == MetricIntent.MULTI_OBJECTIVE:
        if sqa_metric.properties is None:
            sqa_metric.properties = {}
        sqa_metric.properties["skip_runners_and_metrics"] = True


def _get_experiment_sqa_reduced_state(
    experiment_name: str,
    exp_sqa_class: Type[SQAExperiment],
    trial_sqa_class: Type[SQATrial],
    load_trials_in_batches_of_size: Optional[int] = None,
    skip_runners_and_metrics: bool = False,
) -> SQAExperiment:
    """Obtains most of the SQLAlchemy experiment object from DB, with some attributes
    (model state on generator runs, abandoned arms) omitted. Used for loading
    large experiments, in cases where model state history is not required.
    """
    options = get_query_options_to_defer_immutable_duplicates()
    options.append(lazyload("abandoned_arms"))
    options.extend(get_query_options_to_defer_large_model_cols())

    return _get_experiment_sqa(
        experiment_name=experiment_name,
        exp_sqa_class=exp_sqa_class,
        trial_sqa_class=trial_sqa_class,
        trials_query_options=options,
        load_trials_in_batches_of_size=load_trials_in_batches_of_size,
        skip_runners_and_metrics=skip_runners_and_metrics,
    )


def _get_experiment_sqa_immutable_opt_config_and_search_space(
    experiment_name: str,
    exp_sqa_class: Type[SQAExperiment],
    trial_sqa_class: Type[SQATrial],
    load_trials_in_batches_of_size: Optional[int] = None,
    skip_runners_and_metrics: bool = False,
) -> SQAExperiment:
    """For experiments where the search space and opt config are
    immutable, we don't store copies of search space and opt config
    on each generator run. Therefore, there's no need to try to
    load these copies from the DB -- these queries will always return
    an empty list, and are therefore unnecessary and wasteful.
    """
    return _get_experiment_sqa(
        experiment_name=experiment_name,
        exp_sqa_class=exp_sqa_class,
        trial_sqa_class=trial_sqa_class,
        trials_query_options=get_query_options_to_defer_immutable_duplicates(),
        load_trials_in_batches_of_size=load_trials_in_batches_of_size,
        skip_runners_and_metrics=skip_runners_and_metrics,
    )


def _get_experiment_immutable_opt_config_and_search_space(
    experiment_name: str, exp_sqa_class: Type[SQAExperiment]
) -> bool:
    """Return true iff the experiment is designated as having an
    immutable optimization config and search space.
    """
    with session_scope() as session:
        sqa_experiment_properties = (
            session.query(exp_sqa_class.properties)
            .filter_by(name=experiment_name)
            .one_or_none()
        )
        if sqa_experiment_properties is None:
            raise ExperimentNotFoundError(f"Experiment '{experiment_name}' not found.")

    sqa_experiment_properties = sqa_experiment_properties[0] or {}
    return sqa_experiment_properties.get(
        Keys.IMMUTABLE_SEARCH_SPACE_AND_OPT_CONF, False
    )


def _get_experiment_id(experiment_name: str, config: SQAConfig) -> Optional[int]:
    """Get DB ID of the experiment by the given name if its in DB,
    return None otherwise.
    """
    exp_sqa_class = config.class_to_sqa_class[Experiment]
    with session_scope() as session:
        sqa_experiment_id = (
            session.query(exp_sqa_class.id)  # pyre-ignore
            .filter_by(name=experiment_name)
            .one_or_none()
        )

    if sqa_experiment_id is None:
        return None
    return sqa_experiment_id[0]


# ------------------------ Loading `GenerationStrategy`. -----------------------


def load_generation_strategy_by_experiment_name(
    experiment_name: str,
    config: Optional[SQAConfig] = None,
    experiment: Optional[Experiment] = None,
    reduced_state: bool = False,
    skip_runners_and_metrics: bool = False,
) -> GenerationStrategy:
    """Finds a generation strategy attached to an experiment specified by a name
    and restores it from its corresponding SQA object.
    """
    config = config or SQAConfig()
    decoder = Decoder(config=config)
    return _load_generation_strategy_by_experiment_name(
        experiment_name=experiment_name,
        decoder=decoder,
        experiment=experiment,
        reduced_state=reduced_state,
        skip_runners_and_metrics=skip_runners_and_metrics,
    )


def load_generation_strategy_by_id(
    gs_id: int,
    config: Optional[SQAConfig] = None,
    experiment: Optional[Experiment] = None,
    reduced_state: bool = False,
) -> GenerationStrategy:
    """Finds a generation strategy stored by a given ID and restores it."""
    config = config or SQAConfig()
    decoder = Decoder(config=config)
    return _load_generation_strategy_by_id(
        gs_id=gs_id, decoder=decoder, experiment=experiment, reduced_state=reduced_state
    )


def _load_generation_strategy_by_experiment_name(
    experiment_name: str,
    decoder: Decoder,
    experiment: Optional[Experiment] = None,
    reduced_state: bool = False,
    skip_runners_and_metrics: bool = False,
) -> GenerationStrategy:
    """Load a generation strategy attached to an experiment specified by a name,
    using given Decoder instance.

    1) Get SQLAlchemy object from DB.
    2) Convert to corresponding Ax object.
    """
    gs_id = get_generation_strategy_id(experiment_name=experiment_name, decoder=decoder)
    if gs_id is None:
        raise ObjectNotFoundError(
            f"Experiment {experiment_name} does not have a generation strategy "
            "attached to it."
        )
    if not experiment:
        experiment = _load_experiment(
            experiment_name=experiment_name,
            decoder=decoder,
            reduced_state=reduced_state,
            skip_runners_and_metrics=skip_runners_and_metrics,
        )
    return _load_generation_strategy_by_id(
        gs_id=gs_id, decoder=decoder, experiment=experiment, reduced_state=reduced_state
    )


def _load_generation_strategy_by_id(
    gs_id: int,
    decoder: Decoder,
    experiment: Optional[Experiment] = None,
    reduced_state: bool = False,
) -> GenerationStrategy:
    """Finds a generation strategy stored by a given ID and restores it."""
    if reduced_state:
        return decoder.generation_strategy_from_sqa(
            gs_sqa=get_generation_strategy_sqa_reduced_state(
                gs_id=gs_id, decoder=decoder
            ),
            experiment=experiment,
            reduced_state=reduced_state,
        )

    exp_sqa_class = decoder.config.class_to_sqa_class[Experiment]
    immutable_opt_config_and_search_space = (
        (
            _get_experiment_immutable_opt_config_and_search_space(
                experiment_name=experiment.name,
                # pyre-ignore Incompatible parameter type [6]: Expected
                # `Type[SQAExperiment]` for 2nd parameter `exp_sqa_class`
                # to call `_get_experiment_immutable_opt_config_and_search_space`
                # but got `Type[ax.storage.sqa_store.db.SQABase]`.
                exp_sqa_class=exp_sqa_class,
            )
        )
        if experiment is not None
        else False
    )
    if immutable_opt_config_and_search_space:
        gs_sqa = _get_generation_strategy_sqa_immutable_opt_config_and_search_space(
            gs_id=gs_id, decoder=decoder
        )
    else:
        gs_sqa = get_generation_strategy_sqa(gs_id=gs_id, decoder=decoder)

    return decoder.generation_strategy_from_sqa(
        gs_sqa=gs_sqa, experiment=experiment, reduced_state=reduced_state
    )


def get_generation_strategy_id(experiment_name: str, decoder: Decoder) -> Optional[int]:
    """Get DB ID of the generation strategy, associated with the experiment
    with the given name if its in DB, return None otherwise.
    """
    exp_sqa_class = decoder.config.class_to_sqa_class[Experiment]
    gs_sqa_class = decoder.config.class_to_sqa_class[GenerationStrategy]
    with session_scope() as session:
        sqa_gs_id = (
            session.query(gs_sqa_class.id)  # pyre-ignore[16]
            .join(exp_sqa_class.generation_strategy)  # pyre-ignore[16]
            # pyre-fixme[16]: `SQABase` has no attribute `name`.
            .filter(exp_sqa_class.name == experiment_name)
            .one_or_none()
        )

    if sqa_gs_id is None:
        return None
    return sqa_gs_id[0]


def get_generation_strategy_sqa(
    gs_id: int,
    decoder: Decoder,
    # pyre-fixme[2]: Parameter annotation cannot contain `Any`.
    query_options: Optional[List[Any]] = None,
) -> SQAGenerationStrategy:
    """Obtains the SQLAlchemy generation strategy object from DB."""
    gs_sqa_class = cast(
        Type[SQAGenerationStrategy],
        decoder.config.class_to_sqa_class[GenerationStrategy],
    )
    with session_scope() as session:
        query = session.query(gs_sqa_class).filter_by(id=gs_id)
        if query_options:
            query = query.options(*query_options)
        gs_sqa = query.one_or_none()
    if gs_sqa is None:
        raise ObjectNotFoundError(f"Generation strategy with ID #{gs_id} not found.")
    return gs_sqa


def get_generation_strategy_sqa_reduced_state(
    gs_id: int, decoder: Decoder
) -> SQAGenerationStrategy:
    """Obtains most of the SQLAlchemy generation strategy object from DB."""
    gs_sqa_class = cast(
        Type[SQAGenerationStrategy],
        decoder.config.class_to_sqa_class[GenerationStrategy],
    )
    gr_sqa_class = cast(
        Type[SQAGeneratorRun],
        decoder.config.class_to_sqa_class[GeneratorRun],
    )

    gs_sqa = get_generation_strategy_sqa(
        gs_id=gs_id,
        decoder=decoder,
        query_options=[
            lazyload("generator_runs.parameters"),
            lazyload("generator_runs.parameter_constraints"),
            lazyload("generator_runs.metrics"),
            defaultload(gs_sqa_class.generator_runs).defer("model_kwargs"),
            defaultload(gs_sqa_class.generator_runs).defer("bridge_kwargs"),
            defaultload(gs_sqa_class.generator_runs).defer("model_state_after_gen"),
            defaultload(gs_sqa_class.generator_runs).defer("gen_metadata"),
        ],
    )

    # Load full last generator run (including model state), for generation
    # strategy restoration
    if gs_sqa.generator_runs:
        last_generator_run_id = gs_sqa.generator_runs[-1].id
        with session_scope() as session:
            last_gr_sqa = (
                session.query(gr_sqa_class)
                .filter_by(id=last_generator_run_id)
                .one_or_none()
            )

        # Swap last generator run with no state for a generator run with
        # state.
        gs_sqa.generator_runs[len(gs_sqa.generator_runs) - 1] = not_none(last_gr_sqa)

    return gs_sqa


def get_generator_runs_by_id(
    generator_run_ids: List[int],
    decoder: Decoder,
    reduced_state: bool = False,
    immutable_search_space_and_opt_config: bool = False,
) -> List[GeneratorRun]:
    """Bulk fetches generator runs by id."""
    generator_run_sqa_class = decoder.config.class_to_sqa_class[GeneratorRun]
    with session_scope() as session:
        query = session.query(generator_run_sqa_class).filter(
            generator_run_sqa_class.id.in_(generator_run_ids)  # pyre-ignore[16]
        )
        sqa_grs = query.all()
    return [
        decoder.generator_run_from_sqa(
            generator_run_sqa=checked_cast(SQAGeneratorRun, sqa_gr),
            reduced_state=reduced_state,
            immutable_search_space_and_opt_config=immutable_search_space_and_opt_config,
        )
        for sqa_gr in sqa_grs
    ]


def _get_generation_strategy_sqa_immutable_opt_config_and_search_space(
    gs_id: int, decoder: Decoder
) -> SQAGenerationStrategy:
    """Obtains most of the SQLAlchemy generation strategy object from DB."""
    return get_generation_strategy_sqa(
        gs_id=gs_id,
        decoder=decoder,
        query_options=[
            lazyload("generator_runs.parameters"),
            lazyload("generator_runs.parameter_constraints"),
            lazyload("generator_runs.metrics"),
        ],
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.core.parameter_distribution import ParameterDistribution
from ax.core.search_space import RobustSearchSpace
from ax.exceptions.core import UserInputError
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_robust_search_space
from scipy.stats._continuous_distns import norm_gen
from scipy.stats._distn_infrastructure import rv_frozen


class ParameterDistributionTest(TestCase):
    def test_parameter_distribution(self) -> None:
        dist = ParameterDistribution(
            parameters=["x1"],
            distribution_class="norm",
            distribution_parameters={"loc": 0.0, "scale": 1.0},
            multiplicative=True,
        )
        self.assertTrue(dist.multiplicative)
        self.assertIsNone(dist._distribution)
        dist_obj = dist.distribution
        self.assertEqual(dist.parameters, ["x1"])
        self.assertIsInstance(dist_obj, rv_frozen)
        self.assertIsInstance(dist_obj.dist, norm_gen)
        dist_kwds = dist_obj.kwds
        self.assertEqual(dist_kwds["loc"], 0.0)
        self.assertEqual(dist_kwds["scale"], 1.0)

        # Update the dist class.
        self.assertIsNotNone(dist._distribution)
        dist.distribution_class = "cauchy"
        self.assertIsNone(dist._distribution)
        # Update the parameters.
        dist.distribution  # Re-construct the distribution.
        self.assertIsNotNone(dist._distribution)
        dist.distribution_parameters = {"loc": 0.1, "scale": 0.5}
        self.assertIsNone(dist._distribution)

        # Test repr.
        expected_repr = (
            "ParameterDistribution("
            "parameters=['x1'], "
            "distribution_class=cauchy, "
            "distribution_parameters={'loc': 0.1, 'scale': 0.5}, "
            "multiplicative=True)"
        )
        self.assertEqual(str(dist), expected_repr)

        # Test weird distribution name.
        dist = ParameterDistribution(
            parameters=["x1"],
            distribution_class="dummy_dist",
            distribution_parameters={},
        )
        self.assertFalse(dist.multiplicative)
        with self.assertRaises(UserInputError):
            dist.distribution

        # Test `is_environmental`.
        rss = get_robust_search_space()
        self.assertFalse(rss.parameter_distributions[0].is_environmental(rss))
        params = list(rss.parameters.values())
        rss = RobustSearchSpace(
            parameters=params[2:],
            parameter_distributions=rss.parameter_distributions,
            num_samples=4,
            environmental_variables=params[:2],
        )
        self.assertTrue(rss.parameter_distributions[0].is_environmental(rss))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Optional, Union

from ax.core.experiment import Experiment
from ax.modelbridge.generation_strategy import GenerationStrategy

from ax.service.ax_client import AxClient
from ax.service.scheduler import Scheduler
from ax.telemetry.ax_client import AxClientCompletedRecord, AxClientCreatedRecord
from ax.telemetry.common import (
    _get_max_transformed_dimensionality,
    DEFAULT_PRODUCT_SURFACE,
)
from ax.telemetry.experiment import ExperimentCreatedRecord
from ax.telemetry.generation_strategy import GenerationStrategyCreatedRecord
from ax.telemetry.scheduler import SchedulerCompletedRecord, SchedulerCreatedRecord


@dataclass(frozen=True)
class OptimizationCreatedRecord:
    """
    Record of the "Optimization" creation event. This can come from either an
    AxClient or a Scheduler. This Record is especially useful for logging Ax-backed
    optimization results to a tabular database (i.e. one row per Record).
    """

    unique_identifier: str
    owner: str

    # ExperimentCreatedRecord fields
    experiment_name: Optional[str]
    experiment_type: Optional[str]
    num_continuous_range_parameters: int
    num_int_range_parameters_small: int
    num_int_range_parameters_medium: int
    num_int_range_parameters_large: int
    num_log_scale_range_parameters: int
    num_unordered_choice_parameters_small: int
    num_unordered_choice_parameters_medium: int
    num_unordered_choice_parameters_large: int
    num_fixed_parameters: int
    dimensionality: int
    hierarchical_tree_height: int
    num_parameter_constraints: int
    num_objectives: int
    num_tracking_metrics: int
    num_outcome_constraints: int
    num_map_metrics: int
    metric_cls_to_quantity: Dict[str, int]
    runner_cls: str

    # GenerationStrategyCreatedRecord fields
    generation_strategy_name: Optional[str]
    num_requested_initialization_trials: Optional[int]
    num_requested_bayesopt_trials: Optional[int]
    num_requested_other_trials: Optional[int]
    max_parallelism: Optional[int]

    # {AxClient, Scheduler}CreatedRecord fields
    early_stopping_strategy_cls: Optional[str]
    global_stopping_strategy_cls: Optional[str]
    transformed_dimensionality: Optional[int]
    scheduler_total_trials: Optional[int]
    scheduler_max_pending_trials: int
    arms_per_trial: int

    # Top level info
    product_surface: str
    launch_surface: str

    deployed_job_id: int
    trial_evaluation_identifier: Optional[str]

    # Miscellaneous product info
    is_manual_generation_strategy: bool
    warm_started_from: Optional[str]
    num_custom_trials: int
    support_tier: str

    @classmethod
    def from_scheduler(
        cls,
        scheduler: Scheduler,
        unique_identifier: str,
        owner: str,
        product_surface: str,
        launch_surface: str,
        deployed_job_id: int,
        trial_evaluation_identifier: Optional[str],
        is_manual_generation_strategy: bool,
        warm_started_from: Optional[str],
        num_custom_trials: int,
        support_tier: str,
    ) -> OptimizationCreatedRecord:
        scheduler_created_record = SchedulerCreatedRecord.from_scheduler(
            scheduler=scheduler
        )

        experiment_created_record = scheduler_created_record.experiment_created_record
        generation_strategy_created_record = (
            scheduler_created_record.generation_strategy_created_record
        )

        return cls(
            experiment_name=experiment_created_record.experiment_name,
            experiment_type=experiment_created_record.experiment_type,
            num_continuous_range_parameters=(
                experiment_created_record.num_continuous_range_parameters
            ),
            num_int_range_parameters_small=(
                experiment_created_record.num_int_range_parameters_small
            ),
            num_int_range_parameters_medium=(
                experiment_created_record.num_int_range_parameters_medium
            ),
            num_int_range_parameters_large=(
                experiment_created_record.num_int_range_parameters_large
            ),
            num_log_scale_range_parameters=(
                experiment_created_record.num_log_scale_range_parameters
            ),
            num_unordered_choice_parameters_small=(
                experiment_created_record.num_unordered_choice_parameters_small
            ),
            num_unordered_choice_parameters_medium=(
                experiment_created_record.num_unordered_choice_parameters_medium
            ),
            num_unordered_choice_parameters_large=(
                experiment_created_record.num_unordered_choice_parameters_large
            ),
            num_fixed_parameters=experiment_created_record.num_fixed_parameters,
            dimensionality=experiment_created_record.dimensionality,
            hierarchical_tree_height=experiment_created_record.hierarchical_tree_height,
            num_parameter_constraints=(
                experiment_created_record.num_parameter_constraints
            ),
            num_objectives=experiment_created_record.num_objectives,
            num_tracking_metrics=experiment_created_record.num_tracking_metrics,
            num_outcome_constraints=experiment_created_record.num_outcome_constraints,
            num_map_metrics=experiment_created_record.num_map_metrics,
            metric_cls_to_quantity=experiment_created_record.metric_cls_to_quantity,
            runner_cls=experiment_created_record.runner_cls,
            generation_strategy_name=(
                generation_strategy_created_record.generation_strategy_name
            ),
            num_requested_initialization_trials=(
                generation_strategy_created_record.num_requested_initialization_trials
            ),
            num_requested_bayesopt_trials=(
                generation_strategy_created_record.num_requested_bayesopt_trials
            ),
            num_requested_other_trials=(
                generation_strategy_created_record.num_requested_other_trials
            ),
            max_parallelism=generation_strategy_created_record.max_parallelism,
            early_stopping_strategy_cls=(
                scheduler_created_record.early_stopping_strategy_cls
            ),
            global_stopping_strategy_cls=(
                scheduler_created_record.global_stopping_strategy_cls
            ),
            transformed_dimensionality=(
                scheduler_created_record.transformed_dimensionality
            ),
            scheduler_total_trials=scheduler_created_record.scheduler_total_trials,
            scheduler_max_pending_trials=(
                scheduler_created_record.scheduler_max_pending_trials
            ),
            arms_per_trial=scheduler_created_record.arms_per_trial,
            unique_identifier=unique_identifier,
            owner=owner,
            product_surface=product_surface,
            launch_surface=launch_surface,
            deployed_job_id=deployed_job_id,
            trial_evaluation_identifier=trial_evaluation_identifier,
            is_manual_generation_strategy=is_manual_generation_strategy,
            warm_started_from=warm_started_from,
            num_custom_trials=num_custom_trials,
            support_tier=support_tier,
        )

    @classmethod
    def from_ax_client(
        cls,
        ax_client: AxClient,
        unique_identifier: str,
        owner: str,
        product_surface: str,
        launch_surface: str,
        deployed_job_id: int,
        trial_evaluation_identifier: Optional[str],
        is_manual_generation_strategy: bool,
        warm_started_from: Optional[str],
        num_custom_trials: int,
    ) -> OptimizationCreatedRecord:
        ax_client_created_record = AxClientCreatedRecord.from_ax_client(
            ax_client=ax_client
        )

        experiment_created_record = ax_client_created_record.experiment_created_record
        generation_strategy_created_record = (
            ax_client_created_record.generation_strategy_created_record
        )

        return cls(
            experiment_name=experiment_created_record.experiment_name,
            experiment_type=experiment_created_record.experiment_type,
            num_continuous_range_parameters=(
                experiment_created_record.num_continuous_range_parameters
            ),
            num_int_range_parameters_small=(
                experiment_created_record.num_int_range_parameters_small
            ),
            num_int_range_parameters_medium=(
                experiment_created_record.num_int_range_parameters_medium
            ),
            num_int_range_parameters_large=(
                experiment_created_record.num_int_range_parameters_large
            ),
            num_log_scale_range_parameters=(
                experiment_created_record.num_log_scale_range_parameters
            ),
            num_unordered_choice_parameters_small=(
                experiment_created_record.num_unordered_choice_parameters_small
            ),
            num_unordered_choice_parameters_medium=(
                experiment_created_record.num_unordered_choice_parameters_medium
            ),
            num_unordered_choice_parameters_large=(
                experiment_created_record.num_unordered_choice_parameters_large
            ),
            num_fixed_parameters=experiment_created_record.num_fixed_parameters,
            dimensionality=experiment_created_record.dimensionality,
            hierarchical_tree_height=(
                experiment_created_record.hierarchical_tree_height
            ),
            num_parameter_constraints=(
                experiment_created_record.num_parameter_constraints
            ),
            num_objectives=experiment_created_record.num_objectives,
            num_tracking_metrics=experiment_created_record.num_tracking_metrics,
            num_outcome_constraints=experiment_created_record.num_outcome_constraints,
            num_map_metrics=experiment_created_record.num_map_metrics,
            metric_cls_to_quantity=experiment_created_record.metric_cls_to_quantity,
            runner_cls=experiment_created_record.runner_cls,
            generation_strategy_name=(
                generation_strategy_created_record.generation_strategy_name
            ),
            num_requested_initialization_trials=(
                generation_strategy_created_record.num_requested_initialization_trials
            ),
            num_requested_bayesopt_trials=(
                generation_strategy_created_record.num_requested_bayesopt_trials
            ),
            num_requested_other_trials=(
                generation_strategy_created_record.num_requested_other_trials
            ),
            max_parallelism=generation_strategy_created_record.max_parallelism,
            early_stopping_strategy_cls=(
                ax_client_created_record.early_stopping_strategy_cls
            ),
            global_stopping_strategy_cls=(
                ax_client_created_record.global_stopping_strategy_cls
            ),
            transformed_dimensionality=(
                ax_client_created_record.transformed_dimensionality
            ),
            arms_per_trial=ax_client_created_record.arms_per_trial,
            unique_identifier=unique_identifier,
            owner=owner,
            product_surface=product_surface,
            launch_surface=launch_surface,
            deployed_job_id=deployed_job_id,
            trial_evaluation_identifier=trial_evaluation_identifier,
            is_manual_generation_strategy=is_manual_generation_strategy,
            warm_started_from=warm_started_from,
            num_custom_trials=num_custom_trials,
            # The following are not applicable for AxClient
            scheduler_total_trials=None,
            scheduler_max_pending_trials=-1,
            support_tier="",  # This support may be added in the future
        )

    @classmethod
    def from_experiment(
        cls,
        experiment: Experiment,
        generation_strategy: Optional[GenerationStrategy],
        unique_identifier: str,
        owner: str,
        product_surface: str,
        launch_surface: str,
        deployed_job_id: int,
        is_manual_generation_strategy: bool,
        num_custom_trials: int,
        warm_started_from: Optional[str] = None,
        arms_per_trial: Optional[int] = None,
        trial_evaluation_identifier: Optional[str] = None,
    ) -> OptimizationCreatedRecord:
        experiment_created_record = ExperimentCreatedRecord.from_experiment(
            experiment=experiment,
        )
        generation_strategy_created_record = (
            None
            if generation_strategy is None
            else (
                GenerationStrategyCreatedRecord.from_generation_strategy(
                    generation_strategy=generation_strategy,
                )
            )
        )
        arms_per_trial = -1 if arms_per_trial is None else arms_per_trial
        product_surface = (
            DEFAULT_PRODUCT_SURFACE if product_surface is None else product_surface
        )

        num_requested_initialization_trials = (
            None
            if generation_strategy_created_record is None
            else generation_strategy_created_record.num_requested_initialization_trials
        )
        return cls(
            experiment_name=experiment_created_record.experiment_name,
            experiment_type=experiment_created_record.experiment_type,
            num_continuous_range_parameters=(
                experiment_created_record.num_continuous_range_parameters
            ),
            num_int_range_parameters_small=(
                experiment_created_record.num_int_range_parameters_small
            ),
            num_int_range_parameters_medium=(
                experiment_created_record.num_int_range_parameters_medium
            ),
            num_int_range_parameters_large=(
                experiment_created_record.num_int_range_parameters_large
            ),
            num_log_scale_range_parameters=(
                experiment_created_record.num_log_scale_range_parameters
            ),
            num_unordered_choice_parameters_small=(
                experiment_created_record.num_unordered_choice_parameters_small
            ),
            num_unordered_choice_parameters_medium=(
                experiment_created_record.num_unordered_choice_parameters_medium
            ),
            num_unordered_choice_parameters_large=(
                experiment_created_record.num_unordered_choice_parameters_large
            ),
            num_fixed_parameters=experiment_created_record.num_fixed_parameters,
            dimensionality=experiment_created_record.dimensionality,
            hierarchical_tree_height=(
                experiment_created_record.hierarchical_tree_height
            ),
            num_parameter_constraints=(
                experiment_created_record.num_parameter_constraints
            ),
            num_objectives=experiment_created_record.num_objectives,
            num_tracking_metrics=experiment_created_record.num_tracking_metrics,
            num_outcome_constraints=experiment_created_record.num_outcome_constraints,
            num_map_metrics=experiment_created_record.num_map_metrics,
            metric_cls_to_quantity=experiment_created_record.metric_cls_to_quantity,
            runner_cls=experiment_created_record.runner_cls,
            generation_strategy_name=(
                None
                if generation_strategy_created_record is None
                else generation_strategy_created_record.generation_strategy_name
            ),
            num_requested_initialization_trials=num_requested_initialization_trials,
            num_requested_bayesopt_trials=(
                None
                if generation_strategy_created_record is None
                else generation_strategy_created_record.num_requested_bayesopt_trials
            ),
            num_requested_other_trials=(
                None
                if generation_strategy_created_record is None
                else generation_strategy_created_record.num_requested_other_trials
            ),
            max_parallelism=(
                None
                if generation_strategy_created_record is None
                else generation_strategy_created_record.max_parallelism
            ),
            early_stopping_strategy_cls=None,
            global_stopping_strategy_cls=None,
            transformed_dimensionality=(
                None
                if generation_strategy is None
                else _get_max_transformed_dimensionality(
                    search_space=experiment.search_space,
                    generation_strategy=generation_strategy,
                )
            ),
            arms_per_trial=arms_per_trial,
            unique_identifier=unique_identifier,
            owner=owner,
            product_surface=product_surface,
            launch_surface=launch_surface,
            deployed_job_id=deployed_job_id,
            trial_evaluation_identifier=trial_evaluation_identifier,
            is_manual_generation_strategy=is_manual_generation_strategy,
            warm_started_from=warm_started_from,
            num_custom_trials=num_custom_trials,
            # The following are not applicable for AxClient
            scheduler_total_trials=None,
            scheduler_max_pending_trials=-1,
            support_tier="",  # This support may be added in the future
        )


@dataclass(frozen=True)
class OptimizationCompletedRecord:
    """
    Record of the "Optimization" completion event. This can come from either an
    AxClient or a Scheduler. This Record is especially useful for logging Ax-backed
    optimization results to a tabular database (i.e. one row per Record)
    """

    unique_identifier: str

    # ExperimentCompletedRecord fields
    num_initialization_trials: int
    num_bayesopt_trials: int
    num_other_trials: int

    num_completed_trials: int
    num_failed_trials: int
    num_abandoned_trials: int
    num_early_stopped_trials: int

    total_fit_time: int
    total_gen_time: int

    # SchedulerCompletedRecord fields
    best_point_quality: float
    model_fit_quality: float
    model_std_quality: float
    model_fit_generalization: float
    model_std_generalization: float

    improvement_over_baseline: float

    num_metric_fetch_e_encountered: int
    num_trials_bad_due_to_err: int

    # TODO[mpolson64] Deprecate this field as it is redundant with unique_identifier
    deployed_job_id: Optional[int]

    # Miscellaneous deployment specific info
    estimated_early_stopping_savings: float
    estimated_global_stopping_savings: float

    @classmethod
    def from_scheduler(
        cls,
        scheduler: Scheduler,
        unique_identifier: str,
        deployed_job_id: Optional[int],
        estimated_early_stopping_savings: float,
        estimated_global_stopping_savings: float,
    ) -> OptimizationCompletedRecord:
        scheduler_completed_record = SchedulerCompletedRecord.from_scheduler(
            scheduler=scheduler
        )
        experiment_completed_record = (
            scheduler_completed_record.experiment_completed_record
        )

        return cls(
            num_initialization_trials=(
                experiment_completed_record.num_initialization_trials
            ),
            num_bayesopt_trials=experiment_completed_record.num_bayesopt_trials,
            num_other_trials=experiment_completed_record.num_other_trials,
            num_completed_trials=experiment_completed_record.num_completed_trials,
            num_failed_trials=experiment_completed_record.num_failed_trials,
            num_abandoned_trials=experiment_completed_record.num_abandoned_trials,
            num_early_stopped_trials=(
                experiment_completed_record.num_early_stopped_trials
            ),
            total_fit_time=experiment_completed_record.total_fit_time,
            total_gen_time=experiment_completed_record.total_gen_time,
            best_point_quality=scheduler_completed_record.best_point_quality,
            **_extract_model_fit_dict(scheduler_completed_record),
            improvement_over_baseline=(
                scheduler_completed_record.improvement_over_baseline
            ),
            num_metric_fetch_e_encountered=(
                scheduler_completed_record.num_metric_fetch_e_encountered
            ),
            num_trials_bad_due_to_err=(
                scheduler_completed_record.num_trials_bad_due_to_err
            ),
            unique_identifier=unique_identifier,
            deployed_job_id=deployed_job_id,
            estimated_early_stopping_savings=estimated_early_stopping_savings,
            estimated_global_stopping_savings=estimated_global_stopping_savings,
        )

    @classmethod
    def from_ax_client(
        cls,
        ax_client: AxClient,
        unique_identifier: str,
        deployed_job_id: Optional[int],
        estimated_early_stopping_savings: float,
        estimated_global_stopping_savings: float,
    ) -> OptimizationCompletedRecord:
        ax_client_completed_record = AxClientCompletedRecord.from_ax_client(
            ax_client=ax_client
        )
        experiment_completed_record = (
            ax_client_completed_record.experiment_completed_record
        )

        return cls(
            num_initialization_trials=(
                experiment_completed_record.num_initialization_trials
            ),
            num_bayesopt_trials=experiment_completed_record.num_bayesopt_trials,
            num_other_trials=experiment_completed_record.num_other_trials,
            num_completed_trials=experiment_completed_record.num_completed_trials,
            num_failed_trials=experiment_completed_record.num_failed_trials,
            num_abandoned_trials=experiment_completed_record.num_abandoned_trials,
            num_early_stopped_trials=(
                experiment_completed_record.num_early_stopped_trials
            ),
            total_fit_time=experiment_completed_record.total_fit_time,
            total_gen_time=experiment_completed_record.total_gen_time,
            best_point_quality=ax_client_completed_record.best_point_quality,
            **_extract_model_fit_dict(ax_client_completed_record),
            unique_identifier=unique_identifier,
            deployed_job_id=deployed_job_id,
            estimated_early_stopping_savings=estimated_early_stopping_savings,
            estimated_global_stopping_savings=estimated_global_stopping_savings,
            # The following are not applicable for AxClient
            improvement_over_baseline=float("nan"),
            num_metric_fetch_e_encountered=-1,
            num_trials_bad_due_to_err=-1,
        )


def _extract_model_fit_dict(
    completed_record: Union[SchedulerCompletedRecord, AxClientCompletedRecord],
) -> Dict[str, float]:
    model_fit_names = [
        "model_fit_quality",
        "model_std_quality",
        "model_fit_generalization",
        "model_std_generalization",
    ]
    return {n: getattr(completed_record, n) for n in model_fit_names}



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest.mock import Mock, patch

from ax.core.base_trial import TrialStatus
from ax.utils.common.testutils import TestCase
from ax.utils.testing.backend_simulator import BackendSimulator, BackendSimulatorOptions


class BackendSimulatorTest(TestCase):
    @patch("ax.utils.testing.backend_simulator.time.time")
    def test_backend_simulator(self, time_mock: Mock) -> None:
        time_mock.return_value = 0.0
        dt = 0.001
        options = BackendSimulatorOptions(max_concurrency=2)

        # test init
        sim = BackendSimulator(options=options)
        self.assertEqual(sim.max_concurrency, 2)
        self.assertEqual(sim.time_scaling, 1.0)
        self.assertEqual(sim.failure_rate, 0.0)
        self.assertEqual(sim.num_queued, 0)
        self.assertEqual(sim.num_running, 0)
        self.assertEqual(sim.num_failed, 0)
        self.assertEqual(sim.num_completed, 0)

        # test run trial
        sim.run_trial(0, dt)
        self.assertEqual(sim.num_queued, 0)
        self.assertEqual(sim.num_running, 1)
        sim.run_trial(1, dt)
        self.assertEqual(sim.num_queued, 0)
        self.assertEqual(sim.num_running, 2)
        sim.run_trial(2, dt)
        self.assertEqual(sim.num_queued, 1)
        self.assertEqual(sim.num_running, 2)
        status = sim.status()
        self.assertEqual(status.queued, [2])
        self.assertEqual(status.running, [0, 1])
        self.assertEqual(status.failed, [])
        self.assertEqual(status.completed, [])

        # "Wait" some time
        time_mock.return_value += 1.5 * dt
        sim.update()
        self.assertEqual(sim.num_queued, 0)
        self.assertEqual(sim.num_running, 1)
        self.assertEqual(sim.num_failed, 0)
        self.assertEqual(sim.num_completed, 2)

        # extract state for later use
        state = sim.state()

        # let time pass and update
        time_mock.return_value += 10 * dt
        sim.update()
        self.assertEqual(sim.num_queued, 0)
        self.assertEqual(sim.num_running, 0)
        self.assertEqual(sim.num_failed, 0)
        self.assertEqual(sim.num_completed, 3)

        # test reset
        sim.max_concurrency = 3
        sim.time_scaling = 2.0
        sim.failure_rate, 0.5
        sim.reset()
        self.assertEqual(sim.max_concurrency, 2)
        self.assertEqual(sim.time_scaling, 1.0)
        self.assertEqual(sim.failure_rate, 0.0)
        self.assertEqual(sim.num_queued, 0)
        self.assertEqual(sim.num_running, 0)
        self.assertEqual(sim.num_failed, 0)
        self.assertEqual(sim.num_completed, 0)

        # test load state
        sim2 = BackendSimulator.from_state(state)
        self.assertEqual(sim2.max_concurrency, 2)
        self.assertEqual(sim2.time_scaling, 1.0)
        self.assertEqual(sim2.failure_rate, 0.0)
        self.assertEqual(sim2.num_queued, 0)
        self.assertEqual(sim2.num_running, 1)
        self.assertEqual(sim2.num_failed, 0)
        self.assertEqual(sim2.num_completed, 2)
        sim2.update()
        self.assertEqual(sim2.num_queued, 0)
        self.assertEqual(sim2.num_running, 0)
        self.assertEqual(sim2.num_failed, 0)
        self.assertEqual(sim2.num_completed, 3)

        # test failure rate
        options = BackendSimulatorOptions(max_concurrency=2, failure_rate=1.0)
        sim3 = BackendSimulator(options=options)
        sim3.run_trial(0, dt)
        self.assertEqual(sim3.num_queued, 0)
        self.assertEqual(sim3.num_running, 0)
        self.assertEqual(sim3.num_failed, 1)
        self.assertEqual(sim3.num_completed, 0)

    def test_backend_simulator_internal_clock(self) -> None:
        options = BackendSimulatorOptions(
            internal_clock=0.0, use_update_as_start_time=True, max_concurrency=2
        )
        sim = BackendSimulator(options=options)
        sim.run_trial(0, 2)
        sim.run_trial(1, 1)
        sim.run_trial(2, 10)
        self.assertEqual(len(sim.all_trials), 3)
        self.assertEqual(sim.time, 0.0)
        self.assertEqual(sim.num_queued, 1)
        self.assertEqual(
            sim.lookup_trial_index_status(trial_index=0), TrialStatus.RUNNING
        )
        self.assertEqual(
            sim.lookup_trial_index_status(trial_index=1), TrialStatus.RUNNING
        )
        self.assertEqual(
            sim.lookup_trial_index_status(trial_index=2), TrialStatus.STAGED
        )

        sim.update()
        self.assertEqual(sim.num_completed, 1)
        self.assertEqual(sim.num_running, 2)
        self.assertEqual(
            sim.lookup_trial_index_status(trial_index=1), TrialStatus.COMPLETED
        )

        sim.update()
        self.assertEqual(sim.num_completed, 2)
        self.assertEqual(sim.num_running, 1)
        self.assertEqual(
            sim.lookup_trial_index_status(trial_index=0), TrialStatus.COMPLETED
        )

        sim.stop_trial(trial_index=2)
        sim.update()
        self.assertEqual(
            sim.lookup_trial_index_status(trial_index=2), TrialStatus.COMPLETED
        )
        self.assertEqual(
            # pyre-fixme[16]: Optional type has no attribute `sim_completed_time`.
            sim.get_sim_trial_by_index(trial_index=2).sim_completed_time,
            2.0,
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import numpy as np
from ax.models.discrete_base import DiscreteModel
from ax.utils.common.testutils import TestCase


class DiscreteModelTest(TestCase):
    def test_discrete_model_get_state(self) -> None:
        discrete_model = DiscreteModel()
        self.assertEqual(discrete_model._get_state(), {})

    def test_discrete_model_feature_importances(self) -> None:
        discrete_model = DiscreteModel()
        with self.assertRaises(NotImplementedError):
            discrete_model.feature_importances()

    def test_DiscreteModelFit(self) -> None:
        discrete_model = DiscreteModel()
        discrete_model.fit(
            Xs=[[[0]]],
            Ys=[[0]],
            Yvars=[[1]],
            parameter_values=[[0, 1]],
            outcome_names=[],
        )

    def test_discreteModelPredict(self) -> None:
        discrete_model = DiscreteModel()
        with self.assertRaises(NotImplementedError):
            discrete_model.predict([[0]])

    def test_discreteModelGen(self) -> None:
        discrete_model = DiscreteModel()
        with self.assertRaises(NotImplementedError):
            discrete_model.gen(
                n=1, parameter_values=[[0, 1]], objective_weights=np.array([1])
            )

    def test_discreteModelCrossValidate(self) -> None:
        discrete_model = DiscreteModel()
        with self.assertRaises(NotImplementedError):
            discrete_model.cross_validate(
                Xs_train=[[[0]]], Ys_train=[[1]], Yvars_train=[[1]], X_test=[[1]]
            )
